from fontTools.misc.transform import Transform
from fontTools.pens.qtPen import QtPen
from fontTools.ufoLib.pointPen import PointToSegmentPen
from PyQt5.QtCore import Qt

from defconQt.representationFactories.glyphViewFactory import OnlyComponentsQtPen


def _reverseEnumerate(seq):
    n = len(seq)
    for obj in reversed(seq):
        n -= 1
        yield n, obj


# -------------------
# selected components
# -------------------


def SelectedComponentsQPainterPathFactory(glyph):
    pen = OnlyComponentsQtPen(glyph.layer)
    pointPen = PointToSegmentPen(pen)
    selectedPen = OnlyComponentsQtPen(glyph.layer)
    selectedPointPen = PointToSegmentPen(selectedPen)
    originPts = []
    for component in glyph.components:
        if component.selected:
            component.drawPoints(selectedPointPen)
            t = Transform(*component.transformation)
            originPts.append(t.transformPoint((0, 0)))
        else:
            component.drawPoints(pointPen)
    pen.path.setFillRule(Qt.WindingFill)
    selectedPen.path.setFillRule(Qt.WindingFill)
    return (pen.path, selectedPen.path, originPts)


# --------------
# component path
# --------------


def ComponentQPainterPathFactory(component):
    pen = OnlyComponentsQtPen(component.font)
    component.draw(pen)
    pen.path.setFillRule(Qt.WindingFill)
    return pen.path


# ---------------
# selection glyph
# ---------------


def FilterSelectionFactory(glyph):
    copyGlyph = glyph.__class__()
    # points
    pen = copyGlyph.getPointPen()
    for contour in glyph:
        onCurvesSelected = True
        for point in contour:
            if point.segmentType and not point.selected:
                onCurvesSelected = False
                break
        if onCurvesSelected:
            contour.drawPoints(pen)
        else:
            segments = contour.segments
            # put start point at the beginning of a subcontour
            lastSubcontour = None
            for index, segment in reversed(list(enumerate(segments))):
                if segment[-1].selected:
                    lastSubcontour = index
                else:
                    if lastSubcontour is not None:
                        break
            if lastSubcontour is None:
                continue
            segments = segments[lastSubcontour:] + segments[:lastSubcontour]
            # now draw filtered
            shouldMoveTo = True
            for segment in segments:
                on = segment[-1]
                if not on.selected:
                    if not shouldMoveTo:
                        pen.endPath()
                        shouldMoveTo = True
                    continue
                if on.segmentType == "move" and not shouldMoveTo:
                    pen.endPath()
                    shouldMoveTo = True
                if shouldMoveTo:
                    pen.beginPath()
                    pen.addPoint(
                        (on.x, on.y), segmentType="move", smooth=on.smooth, name=on.name
                    )
                    shouldMoveTo = False
                    continue
                for point in segment:
                    pen.addPoint(
                        (point.x, point.y),
                        segmentType=point.segmentType,
                        smooth=point.smooth,
                        name=point.name,
                    )
            if not shouldMoveTo:
                pen.endPath()
    # other stuff
    for component in glyph.components:
        if component.selected:
            component.drawPoints(pen)
    for anchor in glyph.anchors:
        if anchor.selected:
            copyGlyph.appendAnchor(dict(anchor))
    for guideline in glyph.guidelines:
        if guideline.selected:
            copyGlyph.appendGuideline(dict(guideline))
    if glyph.image is not None:
        if glyph.image.selected:
            copyGlyph.image = dict(glyph.image)
    return copyGlyph


def SelectedContoursQPainterPathFactory(glyph):
    copyGlyph = glyph.getRepresentation("TruFont.FilterSelection")
    path = copyGlyph.getRepresentation("defconQt.NoComponentsQPainterPath")
    return path


# --------------------
# curve path and lines
# --------------------


def SplitLinesQPainterPathFactory(glyph):
    pen = SplitLinesFromPathQtPen(glyph.layer)
    for contour in glyph:
        contour.draw(pen)
    pen.path.setFillRule(Qt.WindingFill)
    return (pen.path, pen.lines)


class SplitLinesFromPathQtPen(QtPen):
    def __init__(self, glyphSet, path=None):
        super().__init__(glyphSet, path)
        self.lines = []
        self._curPos = (0, 0)
        self._initPos = None

    def _registerPoint(self, p):
        self._curPos = (p[0], p[1])
        if self._initPos is None:
            self._initPos = self._curPos

    def _moveTo(self, p):
        super()._moveTo(p)
        self._registerPoint(p)

    def _lineTo(self, p):
        self.lines.append((self._curPos[0], self._curPos[1], p[0], p[1]))
        self._moveTo(p)

    def _curveToOne(self, p1, p2, p3):
        super()._curveToOne(p1, p2, p3)
        self._registerPoint(p3)

    def _qCurveToOne(self, p1, p2):
        super()._qCurveToOne(p1, p2)
        self._registerPoint(p2)

    def _closePath(self):
        if self._initPos is not None and self._curPos != self._initPos:
            self.lines.append(
                (self._curPos[0], self._curPos[1], self._initPos[0], self._initPos[1])
            )
        self._initPos = None

    def _endPath(self):
        self._initPos = None



import unicodedata

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QColor, QKeySequence, QPainterPath
from PyQt5.QtWidgets import QApplication

from trufont.drawingTools.baseTool import BaseTool

_path = QPainterPath()
_path.moveTo(5.29, 17.96)
_path.lineTo(6.94, 17.96)
_path.lineTo(7.36, 21.5)
_path.lineTo(7.78, 21.95)
_path.lineTo(12.4, 21.95)
_path.lineTo(12.4, 6.35)
_path.lineTo(11.86, 5.93)
_path.lineTo(9.7, 5.78)
_path.lineTo(9.7, 4.45)
_path.lineTo(18.3, 4.45)
_path.lineTo(18.3, 5.78)
_path.lineTo(16.14, 5.93)
_path.lineTo(15.6, 6.35)
_path.lineTo(15.6, 21.95)
_path.lineTo(20.22, 21.95)
_path.lineTo(20.64, 21.5)
_path.lineTo(21.06, 17.96)
_path.lineTo(22.71, 17.96)
_path.lineTo(22.71, 23.58)
_path.lineTo(5.29, 23.58)
_path.closeSubpath()


def _isUnicodeChar(text):
    return len(text) and unicodedata.category(text) != "Cc"


# XXX: rewind the shaped string when metrics change/anchors are moved


class TextTool(BaseTool):
    icon = _path
    name = QApplication.translate("TextTool", "Text")
    shortcut = "T"
    grabKeyboard = True

    def __init__(self, parent=None):
        super().__init__(parent)

    @property
    def _layoutManager(self):
        return self.parent().layoutManager()

    # TODO: we might want to fold this into LayoutManager
    def _insertUnicodings(self, text):
        unicodeData = self._font.unicodeData
        for c in text:
            glyphName = unicodeData.glyphNameForUnicode(ord(c))
            if glyphName is not None:
                self._layoutManager.insert(glyphName)

    # methods

    def toolActivated(self):
        widget = self.parent()
        # XXX: don't disable tool on setGlyphs, then uncomment this
        # self._layoutManager.initCaret()
        widget.update()

    def toolDisabled(self):
        self.parent().update()

    def drawingAttribute(self, attr, flags):
        if flags.isActiveLayer:
            return attr in ("showGlyphFill", "showGlyphComponentFill")
        return False

    def drawingColor(self, attr, flags):
        if attr == "componentFillColor":
            return Qt.black
        return None

    # events

    def keyPressEvent(self, event):
        key = event.key()
        if event.matches(QKeySequence.Paste):
            # XXX: the menu item should also go down this codepath
            clipboard = QApplication.clipboard()
            mimeData = clipboard.mimeData()
            if mimeData.hasText():
                self._insertUnicodings(mimeData.text())
        elif key == Qt.Key_Left:
            # TODO: we'll probably need to reform this stuff for RTL
            self._layoutManager.caretPrevious()
        elif key == Qt.Key_Right:
            self._layoutManager.caretNext()
        elif key in (Qt.Key_Backspace, Qt.Key_Delete):
            self._layoutManager.delete(forward=(key == Qt.Key_Delete))
        else:
            text = event.text()
            if not _isUnicodeChar(text):
                return
            # text should be just one codepoint, but be safe
            self._insertUnicodings(text)

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            pos = self.parent().mapFromCanvas(event.localPos())
            self._layoutManager.setCaretFromPos(pos)
        else:
            super().mousePressEvent(event)

    def paintBackground(self, painter, index):
        # XXX: we can't currently draw the caret when there's no
        # glyph on canvas
        offset = self._layoutManager.drawingOffset(index)
        if offset is None:
            return
        dx, dy = offset
        bottom, top = self.parent().verticalBounds()
        upm = top - bottom

        painter.save()
        pen = painter.pen()
        pen.setColor(QColor(90, 90, 90))
        pen.setWidth(0)
        painter.setPen(pen)
        painter.translate(dx, bottom + dy)
        painter.drawLine(-30, -25, 0, 0)
        painter.drawLine(0, 0, 30, -25)
        painter.drawLine(0, 0, 0, upm)
        painter.drawLine(-30, upm + 25, 0, upm)
        painter.drawLine(0, upm, 30, upm + 25)
        painter.restore()



from PyQt5.QtCore import QRect, QSize, Qt, pyqtSignal
from PyQt5.QtGui import QColor, QPainter, QPainterPath
from PyQt5.QtWidgets import QApplication, QSizePolicy, QWidget

_hPad = 10
_vPad = 6


class RoundedButtonSet(QWidget):
    clicked = pyqtSignal()
    SingleSelection = 1
    OneOrMoreSelection = 2

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)

        self._options = []
        self._selection = set()
        self._selectionMode = RoundedButtonSet.SingleSelection

    def options(self):
        return self._options

    def setOptions(self, options, selectFirst=True):
        self._options = options
        if selectFirst and self._options:
            self._selection = {0}
        else:
            self._selection = set()
        self.update()

    def selectedOptions(self):
        return [self._options[index] for index in sorted(self._selection)]

    def setSelectedOptions(self, options):
        self._selection = set()
        for option in options:
            index = self._options.index(option)
            self._selection.add(index)
        self.update()

    def selectionMode(self):
        return self._selectionMode

    def setSelectionMode(self, mode):
        self._selectionMode = mode

    # ----------
    # Qt methods
    # ----------

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            for recordIndex, rect in self._optionsRects.items():
                if QRect(*rect).contains(event.pos()):
                    self._clickedIndex = recordIndex
                    self._oldSelection = self._selection
                    self._selection = {recordIndex}
                    if (
                        self._selectionMode > 1
                        and QApplication.keyboardModifiers() & Qt.ShiftModifier
                    ):
                        shiftSelection = self._selection ^ self._oldSelection
                        if shiftSelection:
                            self._selection = shiftSelection
                    else:
                        self._selection |= self._oldSelection
                    break
            self.update()
        else:
            super().mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        if event.button() == Qt.LeftButton:
            clickedRect = self._optionsRects[self._clickedIndex]
            if QRect(*clickedRect).contains(event.pos()):
                self._selection = {self._clickedIndex}
                if (
                    self._selectionMode > 1
                    and QApplication.keyboardModifiers() & Qt.ShiftModifier
                ):
                    shiftSelection = self._selection ^ self._oldSelection
                    if shiftSelection:
                        self._selection = shiftSelection
                self.clicked.emit()
            else:
                self._selection = self._oldSelection
            self.update()
            del self._clickedIndex
            del self._oldSelection
        else:
            super().mouseReleaseEvent(event)

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        self._optionsRects = {}
        w, h = self.width(), self.height()
        metrics = self.fontMetrics()
        hphp = 2 * _hPad

        painter.save()
        path = QPainterPath()
        path.addRoundedRect(0.5, 0.5, w - 1, h - 1, 4, 4)
        painter.fillPath(path, QColor(250, 250, 250))
        x = 0
        linePath = QPainterPath()
        for text in self._options[:-1]:
            x += hphp + metrics.width(text)
            linePath.moveTo(x, 0)
            linePath.lineTo(x, h)
        pen = painter.pen()
        pen.setColor(QColor(218, 218, 218))
        pen.setWidth(0)
        painter.setPen(pen)
        painter.drawPath(path)
        painter.setRenderHint(QPainter.Antialiasing, False)
        painter.drawPath(linePath)
        painter.restore()

        painter.translate(_hPad, _vPad + metrics.ascent())
        left = 0
        for index, text in enumerate(self._options):
            if index in self._selection:
                color = QColor(20, 146, 230)
            else:
                color = QColor(63, 63, 63)
            painter.setPen(color)
            painter.drawText(0, 0, text)
            textWidth = metrics.width(text)
            rectWidth = textWidth + hphp
            rect = (left, 0, rectWidth, h)
            self._optionsRects[index] = rect
            painter.translate(rectWidth, 0)
            left += rectWidth

    def sizeHint(self):
        metrics = self.fontMetrics()
        hphp = 2 * _hPad
        width = sum(metrics.width(text) + hphp for text in self._options) or hphp
        height = 2 * _vPad + metrics.lineSpacing()
        return QSize(width, height)



"""
The *colorVignette* submodule
-----------------------------
"""

from PyQt5.QtCore import QSize, Qt, pyqtSignal
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import (
    QColorDialog,
    QSizePolicy,
    QStyle,
    QStyleOptionFrame,
    QStylePainter,
    QWidget,
)

strikeColor = QColor(144, 60, 74)


class ColorVignette(QWidget):
    """
    A QWidget_ that presents a color in the form of a vignette. Opens up a
    color picker dialog upon double click (or Return key with focus set),
    unless *readOnly* is set to True.
    If *mayClearColor* is True, the color may be cleared with Alt-click (or
    Alt-Return with focus).

    Inspired by ColorPreview and ColorSelector by Mattia Basaglia.

    .. _QWidget: http://doc.qt.io/qt-5/qwidget.html
    """

    colorChanged = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setFocusPolicy(Qt.StrongFocus)
        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        self._color = None
        self._margins = (0, 2, 0, -2)
        self._mayClearColor = True
        self._readOnly = False

    def pickColor(self):
        """
        Opens up a modal QColorDialog_ to choose a new color, and stores it in
        this widget if the user clicks on the Ok button.
        Emits *colorChanged* when successful.

        .. _QColorDialog: http://doc.qt.io/qt-5/qcolordialog.html
        """
        if self._readOnly:
            return
        dialog = QColorDialog(self._color)
        dialog.setOptions(QColorDialog.ShowAlphaChannel)
        ok = dialog.exec_()
        if ok:
            self.setColor(dialog.currentColor())
            self.colorChanged.emit()

    # attributes

    def color(self):
        """
        Returns the color displayed by this widget, or None if no color
        is displayed.
        """
        return self._color

    def setColor(self, color):
        """
        Sets this widget’s QColor_ to *color*.

        The default is None.

        .. _QColor: http://doc.qt.io/qt-5/qcolor.html
        """
        self._color = color
        self.update()

    def margins(self):
        """
        Returns the margins of this widget’s color rectangle as a tuple of four
        scalars.
        """
        dx1, dy1, dx2, dy2 = self._margins
        return (dx1, dy1, -dx2, -dy2)

    def setMargins(self, left, top, right, bottom):
        """
        Sets the margins of the color rectangle inside this widget to *(left,
        top, right, bottom)*.

        The default is (0, 2, 0, 2).
        """
        self._margins = (left, top, -right, -bottom)
        self.update()

    def mayClearColor(self):
        """
        Returns whether this widget allows *color* be set to None by user
        input.
        """
        return self._mayClearColor

    def setMayClearColor(self, value):
        """
        Sets whether the user can clear the color.
        """
        self._mayClearColor = value

    def readOnly(self):
        """
        Returns whether this widget is read-only.
        """
        return self._readOnly

    def setReadOnly(self, value):
        """
        Sets whether this widget should be read-only. If *value* is true,
        :func:`pickColor` will have no effect.
        """
        self._readOnly = value
        self.update()

    # events

    def keyPressEvent(self, event):
        if event.key() & Qt.Key_Return:
            if self._mayClearColor and event.modifiers() & Qt.AltModifier:
                self.setColor(None)
            else:
                self.pickColor()
        else:
            super().keyPressEvent(event)

    def mousePressEvent(self, event):
        if self._mayClearColor and event.modifiers() & Qt.AltModifier:
            self.setColor(None)
        else:
            super().mousePressEvent(event)

    def mouseDoubleClickEvent(self, event):
        self.pickColor()

    def initStyleOption(self, option):
        option.initFrom(self)
        option.lineWidth = self.style().pixelMetric(
            QStyle.PM_DefaultFrameWidth, option, self
        )
        option.midLineWidth = 0
        option.rect = option.rect.adjusted(*self._margins)
        option.state |= QStyle.State_Sunken
        if self._readOnly:
            option.state |= QStyle.State_ReadOnly

    def paint(self, painter, rect):
        panel = QStyleOptionFrame()
        self.initStyleOption(panel)
        style = self.style()
        # use PE_PanelLineEdit instead of static PE_Frame to have hover/focus
        # animation
        style.drawPrimitive(QStyle.PE_PanelLineEdit, panel, painter, self)
        rect = style.subElementRect(QStyle.SE_FrameContents, panel, self)
        painter.fillRect(rect, Qt.white)
        innerRect = rect.adjusted(2, 2, -2, -2)
        if self._color is not None:
            painter.fillRect(innerRect, self._color)
        else:
            pen = painter.pen()
            pen.setColor(strikeColor)
            pen.setWidthF(1.5)
            painter.setPen(pen)
            painter.setRenderHint(QStylePainter.Antialiasing)
            painter.setClipRect(innerRect)
            bL = innerRect.bottomLeft()
            bL.setY(bL.y() + 0.5)
            tR = innerRect.topRight()
            tR.setY(tR.y() + 1)
            painter.drawLine(bL, tR)

    def paintEvent(self, event):
        painter = QStylePainter(self)
        self.paint(painter, self.geometry())

    def resizeEvent(self, event):
        self.update()

    def sizeHint(self):
        return QSize(24, 24)



"""
The *featureCodeEditor* submodule
---------------------------------

The *featureCodeEditor* submodule provides an Adobe `feature file`_ code editor,
and a corresponding syntax highlighter.

.. _`feature file`: http://www.adobe.com/devnet/opentype/afdko/topic_feature_file_syntax.html
"""

import os

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QColor, QTextCharFormat, QTextCursor

from defconQt.controls.baseCodeEditor import BaseCodeEditor, BaseCodeHighlighter

# TODO: maybe move to tools/.
keywordPatterns = [
    "Ascender",
    "Attach",
    "CapHeight",
    "CaretOffset",
    "CodePageRange",
    "Descender",
    "FontRevision",
    "GlyphClassDef",
    "HorizAxis.BaseScriptList",
    "HorizAxis.BaseTagList",
    "HorizAxis.MinMax",
    "IgnoreBaseGlyphs",
    "IgnoreLigatures",
    "IgnoreMarks",
    "LigatureCaretByDev",
    "LigatureCaretByIndex",
    "LigatureCaretByPos",
    "LineGap",
    "MarkAttachClass",
    "MarkAttachmentType",
    "NULL",
    "Panose",
    "RightToLeft",
    "TypoAscender",
    "TypoDescender",
    "TypoLineGap",
    "UnicodeRange",
    "UseMarkFilteringSet",
    "Vendor",
    "VertAdvanceY",
    "VertAxis.BaseScriptList",
    "VertAxis.BaseTagList",
    "VertAxis.MinMax",
    "VertOriginY",
    "VertTypoAscender",
    "VertTypoDescender",
    "VertTypoLineGap",
    "XHeight",
    "anchorDef",
    "anchor",
    "anonymous",
    "anon",
    "by",
    "contour",
    "cursive",
    "device",
    "enumerate",
    "enum",
    "exclude_dflt",
    "featureNames",
    "feature",
    "from",
    "ignore",
    "include_dflt",
    "include",
    "languagesystem",
    "language",
    "lookupflag",
    "lookup",
    "markClass",
    "mark",
    "nameid",
    "name",
    "parameters",
    "position",
    "pos",
    "required",
    "reversesub",
    "rsub",
    "script",
    "sizemenuname",
    "substitute",
    "subtable",
    "sub",
    "table",
    "useExtension",
    "valueRecordDef",
    "winAscent",
    "winDescent",
]


class FeatureCodeHighlighter(BaseCodeHighlighter):
    def __init__(self, parent=None):
        super().__init__(parent)

        keywordFormat = QTextCharFormat()
        keywordFormat.setForeground(QColor(45, 95, 235))
        self.addRule("\\b(?<!\\\\)(%s)\\b" % ("|".join(keywordPatterns)), keywordFormat)

        singleLineCommentFormat = QTextCharFormat()
        singleLineCommentFormat.setForeground(QColor(112, 128, 144))
        self.addRule("#[^\n]*", singleLineCommentFormat)

        groupFormat = QTextCharFormat()
        groupFormat.setForeground(QColor(255, 27, 147))
        self.addRule("@[A-Za-z0-9_.]+", groupFormat)


class FeatureCodeEditor(BaseCodeEditor):
    """
    An Adobe feature file code editor.

    Inherits from :class:`BaseCodeEditor` and provides additional,
    language-specific autocompletion.

    Dropping an FEA file on the widget will create an include(...) directive
    that calls to that file.

    # TODO: maybe insert end block when typing open block, not on newline
    + symetric autocomplete for {} like in the py editor
    """

    openBlockDelimiter = "{"

    def __init__(self, parent=None):
        super().__init__(parent)
        self.highlighter = FeatureCodeHighlighter(self.document())

    def write(self, features):
        """
        TODO: maybe remove this.
        """
        features.text = self.toPlainText()
        self.document().setModified(False)

    def keyPressEvent(self, event):
        key = event.key()
        if key == Qt.Key_Return:
            cursor = self.textCursor()
            indentLvl = self.findLineIndentLevel()
            newBlock = False

            pos = cursor.position()
            cursor.movePosition(QTextCursor.PreviousCharacter, QTextCursor.KeepAnchor)
            if cursor.selectedText() == self.openBlockDelimiter:
                # We don't add a closing tag if there is text right
                # below with the same indentation level because in
                # that case the user might just be looking to add a
                # new line
                ok = cursor.movePosition(QTextCursor.Down)
                if ok:
                    downIndentLvl = self.findLineIndentLevel(cursor)
                    cursor.select(QTextCursor.LineUnderCursor)
                    if (
                        cursor.selectedText().strip() == ""
                        or downIndentLvl <= indentLvl
                    ):
                        newBlock = True
                    cursor.movePosition(QTextCursor.Up)
                else:
                    newBlock = True
                indentLvl += 1

            if newBlock:
                cursor.select(QTextCursor.LineUnderCursor)
                txt = cursor.selectedText().lstrip(" ").split(" ")
                if len(txt) > 1:
                    if len(txt) < 3 and txt[-1][-1] == self.openBlockDelimiter:
                        feature = txt[-1][:-1]
                    else:
                        feature = txt[1]
                else:
                    feature = None
            cursor.setPosition(pos)

            cursor.beginEditBlock()
            cursor.insertText("\n")
            newLineSpace = "".join(self._indent for _ in range(indentLvl))
            cursor.insertText(newLineSpace)
            if newBlock:
                cursor.insertText("\n")
                newLineSpace = "".join(
                    (newLineSpace[: -len(self._indent)], "} ", feature, ";")
                )
                cursor.insertText(newLineSpace)
                cursor.movePosition(QTextCursor.Up)
                cursor.movePosition(QTextCursor.EndOfLine)
                self.setTextCursor(cursor)
            cursor.endEditBlock()
        else:
            super().keyPressEvent(event)

    def dragEnterEvent(self, event):
        if event.source() != self:
            mimeData = event.mimeData()
            if mimeData.hasUrls():
                urls = mimeData.urls()
                for url in urls:
                    if not url.isLocalFile():
                        continue
                    path = url.toLocalFile()
                    if os.path.splitext(path)[1] == ".fea":
                        event.acceptProposedAction()
                        break
                return
        super().dragEnterEvent(event)

    def dropEvent(self, event):
        if event.source() != self:
            mimeData = event.mimeData()
            if mimeData.hasUrls():
                urls = mimeData.urls()
                feaPaths = []
                for url in urls:
                    if not url.isLocalFile():
                        continue
                    path = url.toLocalFile()
                    if os.path.splitext(path)[1] == ".fea":
                        feaPaths.append(path)
                textCursor = self.cursorForPosition(event.pos())
                textCursor.insertText(
                    "\n".join("include(%s);" % feaPath for feaPath in feaPaths)
                )
                self.setTextCursor(textCursor)
                # HACK: Qt uses cleanup routines in a private namespace
                # called by the superclass. Ugh. We have no choice but to make
                # up an empty event if we want to exit drag state.
                mimeData = mimeData.__class__()
                event = event.__class__(
                    event.posF(),
                    event.possibleActions(),
                    mimeData,
                    event.mouseButtons(),
                    event.keyboardModifiers(),
                )
        super().dropEvent(event)



"""
The *glyphCellView* submodule
-----------------------------

The *glyphCellView* submodule provides a widget that displays a list of Glyph_
in cells with their names drawn inside headers.

.. _Glyph: http://ts-defcon.readthedocs.org/en/ufo3/objects/glyph.html
"""

import time
import unicodedata

from defcon import Glyph
from PyQt5.QtCore import QRectF, QSize, Qt, pyqtSignal
from PyQt5.QtGui import QColor, QCursor, QDrag, QPainter, QPainterPath, QPalette
from PyQt5.QtWidgets import QApplication, QScrollArea, QSizePolicy, QWidget

from defconQt.representationFactories.glyphCellFactory import (
    GlyphCellHeaderHeight,
    GlyphCellMinHeightForHeader,
)
from defconQt.tools import platformSpecific
from defconQt.tools.glyphsMimeData import GlyphsMimeData

backgroundColor = Qt.white
cellGridColor = QColor(190, 190, 190)
insertionPositionColor = QColor.fromRgbF(0.16, 0.3, 0.85, 1)

cacheBustSize = 10


class GlyphCellWidget(QWidget):
    """
    The :class:`GlyphCellWidget` widget displays a list of Glyph_ organized
    in cells with their names inside headers.

    This widget allows keyboard navigation and selection of one or more cells,
    and emits the *glyphActivated* signal upon double-click or Return keyboard
    pressed and yields the concerned Glyph_.
    It also has a *selectionChanged* signal.

    Drag and drop can be enabled with *setAcceptDrops(True)*. This widget
    then supports reordering its glyph cells (in which case it emits
    *orderChanged*).

    # TODO: navigation with Shift is perfectible

    .. _Glyph: http://ts-defcon.readthedocs.org/en/ufo3/objects/glyph.html
    """

    glyphActivated = pyqtSignal(Glyph)
    glyphsDropped = pyqtSignal()
    selectionChanged = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setAttribute(Qt.WA_KeyCompression)
        self.setFocusPolicy(Qt.ClickFocus)
        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        self._cellWidth = 50
        self._cellHeight = 50
        self._cellWidthExtra = 0
        self._cellSizeCache = set()
        self._glyphs = []

        self._inputString = ""
        self._lastKeyInputTime = None
        self._selection = set()
        self._lastSelectedCell = None

        self._currentDropIndex = None
        self._maybeDragPosition = None

        self._columnCount = 0
        self._rowCount = 0

        self._cellRepresentationName = "defconQt.GlyphCell"
        self._cellRepresentationArguments = {}

    # --------------
    # Custom methods
    # --------------

    def scrollArea(self):
        return self._scrollArea

    def setScrollArea(self, scrollArea):
        scrollArea.setWidget(self)
        self._scrollArea = scrollArea

    def _getCurrentRepresentation(self, glyph):
        name = self._cellRepresentationName
        args = self._cellRepresentationArguments
        args["width"] = self._cellWidth + 2 * self._cellWidthExtra
        args["height"] = self._cellHeight
        args["pixelRatio"] = self.devicePixelRatio()

        self._cellSizeCache.add((args["width"], args["height"]))
        return glyph.getRepresentation(name, **args)

    def preloadGlyphCellImages(self):
        """
        This preloads the glyphs’ current cell representations in the defcon
        representations_ cache.

        .. _representations: http://ts-defcon.readthedocs.org/en/ufo3/concepts/representations.html
        """
        for glyph in self._glyphs:
            self._getCurrentRepresentation(glyph)

    def glyphs(self):
        """
        Returns the list of glyphs displayed by this widget.
        """
        return self._glyphs

    def setGlyphs(self, glyphs):
        """
        Sets the list of Glyph_ *glyphs* displayed by this widget.

        The widget will try to preserve selection across the changing glyph
        set.

        The default is [].

        .. _Glyph: http://ts-defcon.readthedocs.org/en/ufo3/objects/glyph.html
        """
        currentSelection = [self._glyphs[index] for index in self._selection]
        newSelection = {
            glyphs.index(glyph) for glyph in currentSelection if glyph in glyphs
        }
        self._glyphs = glyphs
        self.setSelection(newSelection)
        self.adjustSize()

    def glyphsForIndexes(self, indexes):
        """
        Returns a list of glyphs that are at *indexes*.

        Indexes must be in range(len(glyphs)).
        """
        return [self._glyphs[i] for i in indexes]

    def cellSize(self):
        """
        Returns a tuple of *(cellWidth, cellHeight)*.
        """
        return self._cellWidth, self._cellHeight

    def setCellSize(self, width, height=None):
        """
        Sets the cellWidth to *width* and cellHeight to *height*.

        If *height* is None, cellHeight will be assigned *width*.

        The default is 50.
        """
        if height is None:
            height = width
        self._cellWidth = width
        self._cellHeight = height
        self._calculateCellWidthExtra()
        self._checkFlushCache()
        self.adjustSize()

    def cellRepresentationName(self):
        """
        Returns the name of the current glyph cell representation.
        """
        return self._cellRepresentationName

    def setCellRepresentationName(self, name):
        """
        Sets the name of the current glyph cell representation.

        The default is "defconQt.GlyphCell".
        """
        self._cellRepresentationName = name
        self.update()

    def cellRepresentationArguments(self):
        """
        Returns the current arguments passed to the glyph cell
        representationFactory.

        The default is {}.
        """
        return self._cellRepresentationArguments

    def setCellRepresentationArguments(self, kwargs):
        """
        Sets the arguments that will be passed to the glyph cell
        representationFactory.

        *kwargs* should be a dict.
        """
        self._cellRepresentationArguments = kwargs
        self.update()

    # selection

    def selection(self):
        """
        Returns a list of the current selection indexes, in ascending order.
        """
        return sorted(self._selection)

    def setSelection(self, selection, lastSelectedCell=None):
        """
        Sets this widget’s selection to a set of indexes *selection*.

        *lastSelectedCell* can be specified as an index that’s a member of the
        *selection* set, otherwise it will be set to min(selection) or None if
        *selection* is an empty set.

        The default *selection* is set(). The default *lastSelectedCell*
        is None.
        """
        self._selection = set(selection)
        if lastSelectedCell is not None:
            assert lastSelectedCell in self._selection
            self._lastSelectedCell = lastSelectedCell
        else:
            # fallback
            if not selection:
                self._lastSelectedCell = None
            else:
                self._lastSelectedCell = min(self._selection)
        if self._lastSelectedCell is not None:
            self.scrollToCell(self._lastSelectedCell)
        self.selectionChanged.emit()
        self.update()

    def lastSelectedCell(self):
        return self._lastSelectedCell

    def lastSelectedGlyph(self):
        cell = self._lastSelectedCell
        if cell is not None:
            return self._glyphs[cell]
        return None

    def _calculateCellWidthExtra(self):
        if self._columnCount:
            rem = self.width() % self._cellWidth
            self._cellWidthExtra = rem // (2 * self._columnCount)
        else:
            self._cellWidthExtra = 0

    def _checkFlushCache(self):
        if len(self._cellSizeCache) >= cacheBustSize:
            for glyph in self._glyphs:
                glyph.destroyRepresentation(self._cellRepresentationName)
            self._cellSizeCache = set()

    # ----------
    # Qt methods
    # ----------

    def sizeHint(self):
        parent = self.parent()
        width = parent.width() if parent is not None else self.width()
        if self._glyphs:
            columnCount = int(width / self._cellWidth)
            if columnCount == 0:
                columnCount = 1
            if columnCount > len(self._glyphs):
                columnCount = len(self._glyphs)
            rowCount = len(self._glyphs) // columnCount
            if columnCount * rowCount < len(self._glyphs):
                rowCount += 1
            newWidth = self._cellWidth * columnCount
            newHeight = self._cellHeight * rowCount
        else:
            columnCount = 0
            rowCount = 0
            newWidth = newHeight = 0
        self._columnCount = columnCount
        self._rowCount = rowCount
        return QSize(newWidth, newHeight)

    def paintEvent(self, event):
        painter = QPainter(self)
        visibleRect = event.rect()
        columnCount = self._columnCount
        extra = self._cellWidthExtra
        cellWidth, cellHeight = self._cellWidth + 2 * extra, self._cellHeight
        glyphCount = len(self._glyphs)
        if columnCount:
            paintWidth = min(glyphCount, columnCount) * cellWidth
        else:
            paintWidth = 0
        left = 0
        top = cellHeight

        painter.fillRect(visibleRect, Qt.white)
        for index, glyph in enumerate(self._glyphs):
            t = top - cellHeight
            rect = (left, t, cellWidth, cellHeight)

            if visibleRect.intersects(visibleRect.__class__(*rect)):
                if index in self._selection:
                    palette = self.palette()
                    active = palette.currentColorGroup() != QPalette.Inactive
                    opacityMultiplier = platformSpecific.colorOpacityMultiplier()
                    selectionColor = palette.color(QPalette.Highlight)
                    # TODO: alpha values somewhat arbitrary (here and in
                    # glyphLineView)
                    selectionColor.setAlphaF(0.2 * opacityMultiplier if active else 0.7)
                    painter.fillRect(
                        QRectF(left, t, cellWidth, cellHeight), selectionColor
                    )

                pixmap = self._getCurrentRepresentation(glyph)
                painter.drawPixmap(left, t, pixmap)

                # XXX: this hacks around the repr internals
                if (
                    index in self._selection
                    and cellHeight >= GlyphCellMinHeightForHeader
                ):
                    painter.fillRect(
                        QRectF(
                            left,
                            t + cellHeight - GlyphCellHeaderHeight,
                            cellWidth,
                            GlyphCellHeaderHeight,
                        ),
                        selectionColor,
                    )

            left += cellWidth
            if left + cellWidth > paintWidth:
                left = 0
                top += cellHeight

        # drop insertion position
        dropIndex = self._currentDropIndex
        if dropIndex is not None:
            if columnCount:
                x = (dropIndex % columnCount) * cellWidth
                y = (dropIndex // columnCount) * cellHeight
                # special-case the end-column
                if (
                    dropIndex == glyphCount
                    and glyphCount < self.width() // self._cellWidth
                    or self.mapFromGlobal(QCursor.pos()).y() < y
                ):
                    x = columnCount * cellWidth
                    y -= cellHeight
            else:
                x = y = 0
            path = QPainterPath()
            path.addRect(x - 2, y, 3, cellHeight)
            path.addEllipse(x - 5, y - 5, 9, 9)
            path.addEllipse(x - 5, y + cellHeight - 5, 9, 9)
            path.setFillRule(Qt.WindingFill)
            pen = painter.pen()
            pen.setColor(Qt.white)
            pen.setWidth(2)
            painter.setPen(pen)
            painter.setRenderHint(QPainter.Antialiasing)
            painter.drawPath(path)
            painter.fillPath(path, insertionPositionColor)

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self._calculateCellWidthExtra()

    # ---------
    # Selection
    # ---------

    def _linearSelection(self, index):
        if not self._selection:
            newSelection = {index}
        else:
            if index < self._lastSelectedCell:
                newSelection = self._selection | set(
                    range(index, self._lastSelectedCell + 1)
                )
            else:
                newSelection = self._selection | set(
                    range(self._lastSelectedCell, index + 1)
                )
        return newSelection

    def scrollToCell(self, index):
        """
        Scrolls the parent QScrollArea_ to show the glyph cell at *index*, if
        suitable parent widget is present.

        .. _QScrollArea: http://doc.qt.io/qt-5/qscrollarea.html
        """
        cellWidth, cellHeight = self._cellWidth, self._cellHeight
        scrollArea = self._scrollArea

        x = (0.5 + index % self._columnCount) * cellWidth
        y = (0.5 + index // self._columnCount) * cellHeight
        if scrollArea is not None:
            scrollArea.ensureVisible(x, y, cellWidth, cellHeight)

    # mouse

    def mousePressEvent(self, event):
        if event.button() in (Qt.LeftButton, Qt.RightButton):
            self._oldSelection = set(self._selection)
            index = self._findIndexForEvent(event)
            modifiers = event.modifiers()

            if index is None:
                if not (modifiers & Qt.ControlModifier or modifiers & Qt.ShiftModifier):
                    # TODO: consider setSelection(None)
                    self.setSelection(set())
                return

            if modifiers & Qt.ControlModifier:
                if index in self._selection:
                    self._selection.remove(index)
                else:
                    self._selection.add(index)
            elif modifiers & Qt.ShiftModifier:
                self._selection = self._linearSelection(index)
            elif index not in self._selection:
                self._selection = {index}
            else:
                self._maybeDragPosition = event.localPos()
            self._lastSelectedCell = index
            self.selectionChanged.emit()
            self.update()
        else:
            super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        if event.buttons() & (Qt.LeftButton | Qt.RightButton):
            index = self._findIndexForEvent(event, True)
            if index == self._lastSelectedCell:
                return
            if self.maybeExecuteDrag(event):
                return
            self.scrollToCell(index)
            if index >= len(self._glyphs):
                return

            modifiers = event.modifiers()
            if modifiers & Qt.ControlModifier:
                if index in self._selection and index in self._oldSelection:
                    self._selection.remove(index)
                elif index not in self._selection and index not in self._oldSelection:
                    self._selection.add(index)
            elif modifiers & Qt.ShiftModifier:
                self._selection = self._linearSelection(index)
            else:
                self._selection = {index}
            self._lastSelectedCell = index
            self.selectionChanged.emit()
            self.update()
        else:
            super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        if event.button() in (Qt.LeftButton, Qt.RightButton):
            self._maybeDragPosition = None
            # XXX: we should use modifiers registered on click
            if not event.modifiers() & Qt.ShiftModifier:
                if self._lastSelectedCell is not None:
                    self._selection = {self._lastSelectedCell}
                else:
                    self._selection = set()
                self.update()
            self._oldSelection = None
        else:
            super().mouseReleaseEvent(event)

    def mouseDoubleClickEvent(self, event):
        if event.button() in (Qt.LeftButton, Qt.RightButton):
            index = self._findIndexForEvent(event)
            if index is not None:
                self.glyphActivated.emit(self._glyphs[index])
        else:
            super().mouseDoubleClickEvent(event)

    def _findIndexForEvent(self, event, allowAllViewport=False):
        cellHeight, cellWidth = (
            self._cellHeight,
            self._cellWidth + self._cellWidthExtra * 2,
        )
        glyphCount = len(self._glyphs)
        x, y = event.x(), event.y()
        visibleWidth = min(glyphCount, self._columnCount) * cellWidth
        if (
            not allowAllViewport or self._lastSelectedCell is None
        ) and x >= visibleWidth:
            return None
        x = max(0, min(event.x(), visibleWidth - 1))
        y = max(0, min(event.y(), self.height() - 1))
        index = (y // cellHeight) * self._columnCount + x // cellWidth
        if not allowAllViewport and index >= glyphCount:
            return None
        return index

    # key

    def selectAll(self):
        """
        Selects all glyphs displayed by this widget.
        """
        newSelection = set(range(len(self._glyphs)))
        self.setSelection(newSelection)

    def keyPressEvent(self, event):
        key = event.key()
        modifiers = event.modifiers()
        if key in (Qt.Key_Up, Qt.Key_Down, Qt.Key_Left, Qt.Key_Right):
            self._arrowKeyPressEvent(event)
        elif key == Qt.Key_Return:
            index = self._lastSelectedCell
            if index is not None:
                self.glyphActivated.emit(self._glyphs[index])
        elif modifiers in (Qt.NoModifier, Qt.ShiftModifier):
            self._glyphNameInputEvent(event)
        else:
            super().keyPressEvent(event)

    def _arrowKeyPressEvent(self, event):
        count = event.count()
        key = event.key()
        modifiers = event.modifiers()
        # TODO: it might be the case that self._lastSelectedCell cannot be None
        # when we arrive here whatsoever
        if self._lastSelectedCell is not None:
            if key == Qt.Key_Up:
                delta = -self._columnCount
            elif key == Qt.Key_Down:
                delta = self._columnCount
            elif key == Qt.Key_Left:
                delta = -1
            elif key == Qt.Key_Right:
                delta = 1
            newSel = self._lastSelectedCell + delta * count
            if newSel < 0 or newSel >= len(self._glyphs):
                return
            if modifiers & Qt.ShiftModifier:
                self._selection |= self._linearSelection(newSel)
            else:
                self._selection = {newSel}
            self._lastSelectedCell = newSel
            self.scrollToCell(newSel)
            self.selectionChanged.emit()
            self.update()

    def _glyphNameInputEvent(self, event):
        inputText = event.text()
        if not self._isUnicodeChar(inputText):
            return
        # get the current time
        rightNow = time.time()
        # no time defined. define it.
        if self._lastKeyInputTime is None:
            self._lastKeyInputTime = rightNow
        # if the last input was too long ago,
        # clear away the old input
        if rightNow - self._lastKeyInputTime > 0.75:
            self._inputString = ""
        # reset the clock
        self._lastKeyInputTime = rightNow
        self._inputString = self._inputString + inputText

        match = None
        matchIndex = None
        lastResort = None
        lastResortIndex = None
        for index, glyph in enumerate(self._glyphs):
            item = glyph.name
            if item is None:
                continue
            # if the item starts with the input string, it is considered
            # a match
            if item.startswith(self._inputString):
                if match is None:
                    match = item
                    matchIndex = index
                    continue
                # only if the item is less than the previous match is it
                # a more relevant match
                # example:
                # given this order: sys, signal
                # and this input string: s
                # sys will be the first match, but signal is the more
                # accurate match
                if item < match:
                    match = item
                    matchIndex = index
                    continue
            # if the item is greater than the input string,it can be used
            # as a last resort
            # example:
            # given this order: vanilla, zipimport
            # and this input string: x
            # zipimport will be used as the last resort
            if item > self._inputString:
                if lastResort is None:
                    lastResort = item
                    lastResortIndex = index
                    continue
                # if existing the last resort is greater than the item
                # the item is a closer match to the input string
                if lastResort > item:
                    lastResort = item
                    lastResortIndex = index
                    continue

        if matchIndex is not None:
            newSelection = matchIndex
        elif lastResortIndex is not None:
            newSelection = lastResortIndex
        else:
            return
        self.setSelection({newSelection})

    def _isUnicodeChar(self, char):
        return len(char) and unicodedata.category(char) != "Cc"

    # -------------
    # Drag and drop
    # -------------

    def maybeExecuteDrag(self, event):
        if self._maybeDragPosition is None:
            return False
        if (
            event.localPos() - self._maybeDragPosition
        ).manhattanLength() < QApplication.startDragDistance():
            return False

        drag = QDrag(self)
        glyphs = self.glyphsForIndexes(self.selection())
        mimeData = GlyphsMimeData()
        mimeData.setGlyphs(glyphs)
        drag.setMimeData(mimeData)
        drag.exec_()
        self._maybeDragPosition = None
        return True

    def dragEnterEvent(self, event):
        if event.source() == self:
            # glyph reordering
            event.acceptProposedAction()
        else:
            super().dragEnterEvent(event)

    def dragMoveEvent(self, event):
        pos = event.pos()
        dropIndex = int(
            self._columnCount * (pos.y() // self._cellHeight)
            + (pos.x() + 0.5 * self._cellWidth) // self._cellWidth
        )
        self._currentDropIndex = min(dropIndex, len(self._glyphs))
        self.update()

    def dragLeaveEvent(self, event):
        self._currentDropIndex = None
        self.update()

    def dropEvent(self, event):
        insert = self._currentDropIndex
        newGlyphs = event.mimeData().glyphs()
        # put all glyphs to be moved to None (deleting them would
        # invalidate our insert indexes)
        if event.source() == self:
            selection = self._selection
            if selection:
                for index in selection:
                    self._glyphs[index] = None
        # insert newGlyphs into the list
        lst = self._glyphs[:insert] + newGlyphs + self._glyphs[insert:]
        self._glyphs = lst
        # now, elide None
        self._currentDropIndex = None
        self._glyphs = [glyph for glyph in self._glyphs if glyph is not None]
        self.setSelection(set())
        self.glyphsDropped.emit()
        self.update()


class GlyphCellView(QScrollArea):
    """
    The :class:`GlyphCellView` widget is a QScrollArea_ that contains a
    :class:`GlyphCellWidget`.

    It reimplements :class:`GlyphLineWidget` public API.

    Here’s an example that displays the glyphs of a Font_ in a
    :class:`GlyphLineView`:

    >>> from defconQt.controls.glyphCellView import GlyphCellView
    >>> window = BaseWindow()
    >>> glyphs = [font[name] for name in font.keys()]
    >>> view = GlyphCellView(window)
    >>> view.setGlyphs(glyphs)
    >>>
    >>> layout = QVBoxLayout(window)
    >>> layout.addWidget(view)
    >>> window.show()

    TODO: add sample image

    .. _Glyph: http://ts-defcon.readthedocs.org/en/ufo3/objects/glyph.html
    .. _QScrollArea: http://doc.qt.io/qt-5/qscrollarea.html
    """

    glyphCellWidgetClass = GlyphCellWidget

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWidgetResizable(True)

        self._glyphCellWidget = self.glyphCellWidgetClass(self)
        self._glyphCellWidget.setScrollArea(self)
        # reexport signals
        self.glyphActivated = self._glyphCellWidget.glyphActivated
        self.glyphsDropped = self._glyphCellWidget.glyphsDropped
        self.selectionChanged = self._glyphCellWidget.selectionChanged

    # -------------
    # Notifications
    # -------------

    def _subscribeToGlyphs(self, glyphs):
        handledGlyphs = set()
        handledFonts = set()
        for glyph in glyphs:
            if glyph in handledGlyphs:
                continue
            handledGlyphs.add(glyph)
            glyph.addObserver(self, "_glyphChanged", "Glyph.Changed")
            font = glyph.font
            if font is None:
                continue
            if font in handledFonts:
                continue
            handledFonts.add(font)
            font.info.addObserver(self, "_fontChanged", "Info.Changed")

    def _unsubscribeFromGlyphs(self):
        handledGlyphs = set()
        handledFonts = set()
        if self._glyphCellWidget is not None:
            glyphs = self._glyphCellWidget.glyphs()
            for glyph in glyphs:
                if glyph in handledGlyphs:
                    continue
                handledGlyphs.add(glyph)
                glyph.removeObserver(self, "Glyph.Changed")
                font = glyph.font
                if font is None:
                    continue
                if font in handledFonts:
                    continue
                handledFonts.add(font)
                font.info.removeObserver(self, "Info.Changed")

    def _glyphChanged(self, notification):
        self._glyphCellWidget.update()

    def _fontChanged(self, notification):
        info = notification.object
        font = info.font
        glyphs = self._glyphCellWidget.glyphs()
        representationName = self._glyphCellWidget.cellRepresentationName()
        for glyph in glyphs:
            if glyph.font == font:
                glyph.destroyRepresentation(representationName)
        self._glyphCellWidget.update()

    # --------------
    # Public methods
    # --------------

    def preloadGlyphCellImages(self):
        self._glyphCellWidget.preloadGlyphCellImages()

    def glyphs(self):
        return self._glyphCellWidget.glyphs()

    def setGlyphs(self, glyphs):
        self._unsubscribeFromGlyphs()
        self._glyphCellWidget.setGlyphs(glyphs)
        self._subscribeToGlyphs(glyphs)

    def glyphsForIndexes(self, indexes):
        return self._glyphCellWidget.glyphsForIndexes(indexes)

    def cellSize(self):
        return self._glyphCellWidget.cellSize()

    def setCellSize(self, width, height=None):
        self._glyphCellWidget.setCellSize(width, height)

    def cellRepresentationName(self):
        return self._glyphCellWidget.cellRepresentationName()

    def setCellRepresentationName(self, name):
        self._glyphCellWidget.setCellRepresentationName(name)

    def cellRepresentationArguments(self):
        return self._glyphCellWidget.cellRepresentationArguments()

    def setCellRepresentationArguments(self, kwargs):
        self._glyphCellWidget.setCellRepresentationArguments(kwargs)

    def selection(self):
        return self._glyphCellWidget.selection()

    def setSelection(self, selection, lastSelectedCell=None):
        self._glyphCellWidget.setSelection(selection, lastSelectedCell)

    def lastSelectedCell(self):
        return self._glyphCellWidget.lastSelectedCell()

    def lastSelectedGlyph(self):
        return self._glyphCellWidget.lastSelectedGlyph()

    def selectAll(self):
        self._glyphCellWidget.selectAll()

    # ----------
    # Qt methods
    # ----------

    def setAcceptDrops(self, value):
        self._glyphCellWidget.setAcceptDrops(value)



"""
The *glyphView* submodule
-----------------------------

The *glyphView* submodule provides widgets that render a Glyph_, with
various display parameters.

.. _Glyph: http://ts-defcon.readthedocs.org/en/ufo3/objects/glyph.html
"""

from PyQt5.QtCore import QEvent, QPoint, QPointF, QSize, Qt, pyqtSignal
from PyQt5.QtGui import QCursor, QPainter
from PyQt5.QtWidgets import QPinchGesture, QScrollArea, QSizePolicy, QWidget

from defconQt.tools import drawing, platformSpecific

# TODO: when the scrollArea resizes, keep the view centered

GlyphViewMinSizeForDetails = 175

UIFont = platformSpecific.otherUIFont()


class GlyphWidget(QWidget):
    pointSizeModified = pyqtSignal(int)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setContextMenuPolicy(Qt.DefaultContextMenu)
        self.setFocusPolicy(Qt.ClickFocus)
        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        self._fitViewport = True
        self._glyph = None
        self._scrollArea = None

        # drawing attributes
        self._layerDrawingAttributes = {}
        self._fallbackDrawingAttributes = dict(
            showGlyphFill=False,
            showGlyphStroke=True,
            showGlyphOnCurvePoints=True,
            showGlyphStartPoints=True,
            showGlyphOffCurvePoints=True,
            showGlyphPointCoordinates=False,
            showGlyphAnchors=True,
            showGlyphMetrics=True,
            showGlyphGuidelines=True,
            showGlyphImage=True,
            showFontVerticalMetrics=True,
            showFontVerticalMetricsTitles=False,
            showFontGuidelines=True,
            showFontPostscriptBlues=True,
            showFontPostscriptFamilyBlues=False,  # TODO: test appearance of this
            # combined w/ blues
        )

        # cached vertical metrics
        self._unitsPerEm = 1000
        self._descender = -250
        self._capHeight = 750
        self._ascender = 750

        # drawing data cache
        self._drawingRect = None
        self._scale = 1.0
        self._inverseScale = 0.1
        self._impliedPointSize = 1000

        # drawing calculation
        self._noPointSizePadding = 200
        self._verticalCenterYBuffer = 0

        self._backgroundColor = Qt.white

    # --------------
    # Custom Methods
    # --------------

    def drawingRect(self):
        return self._drawingRect

    def inverseScale(self):
        return self._inverseScale

    def pointSize(self):
        return self._impliedPointSize

    def setPointSize(self, pointSize):
        scrollArea = self._scrollArea
        newScale = pointSize / self._unitsPerEm
        if scrollArea is not None:
            # compute new scrollbar position
            hSB = scrollArea.horizontalScrollBar()
            vSB = scrollArea.verticalScrollBar()
            viewport = scrollArea.viewport()
            centerPos = QPoint(viewport.width() / 2, viewport.height() / 2)
            pos = self.mapToCanvas(self.mapFromParent(centerPos))
        self.setScale(newScale)
        if scrollArea is not None:
            pos = self.mapFromCanvas(pos)
            delta = pos - self.mapFromParent(centerPos)
            hSB.setValue(hSB.value() + delta.x())
            vSB.setValue(vSB.value() + delta.y())

    def scale(self):
        return self._scale

    def setScale(self, scale):
        self._scale = scale
        if self._scale <= 0:
            self._scale = 0.01
        self._inverseScale = 1.0 / self._scale
        self._impliedPointSize = self._unitsPerEm * self._scale
        self.adjustSize()

    def glyph(self):
        return self._glyph

    def setGlyph(self, glyph):
        self._glyph = glyph
        self._font = None
        if glyph is not None:
            font = self._font = glyph.font
            if font is not None:
                self._unitsPerEm = font.info.unitsPerEm
                if self._unitsPerEm is None:
                    self._unitsPerEm = 1000
                self._descender = font.info.descender
                if self._descender is None:
                    self._descender = -250
                self._ascender = font.info.ascender
                if self._ascender is None:
                    self._ascender = self._unitsPerEm + self._descender
                self._capHeight = font.info.capHeight
                if self._capHeight is None:
                    self._capHeight = self._ascender
            self.setScale(self._scale)
        self.update()

    def scrollArea(self):
        return self._scrollArea

    def setScrollArea(self, scrollArea):
        scrollArea.setWidget(self)
        self._scrollArea = scrollArea

    # fitting

    def centerOn(self, pos):
        """
        Centers this widget’s *scrollArea* on QPointF_ *pos*.

        .. _QPointF: http://doc.qt.io/qt-5/qpointf.html
        """
        scrollArea = self._scrollArea
        if scrollArea is None:
            return
        hSB = scrollArea.horizontalScrollBar()
        vSB = scrollArea.verticalScrollBar()
        viewport = scrollArea.viewport()
        hValue = hSB.minimum() + hSB.maximum() - (pos.x() - viewport.width() / 2)
        hSB.setValue(hValue)
        vSB.setValue(pos.y() - viewport.height() / 2)

    def _calculateDrawingRect(self):
        # calculate and store the drawing rect
        # TODO: we only need the width here
        glyphWidth = self._getGlyphWidthHeight()[0] * self._scale
        diff = self.width() - glyphWidth
        xOffset = round((diff / 2) * self._inverseScale)

        yOffset = self._verticalCenterYBuffer * self._inverseScale
        yOffset -= self._descender

        w = self.width() * self._inverseScale
        h = self.height() * self._inverseScale
        self._drawingRect = (-xOffset, -yOffset, w, h)

    def _getGlyphWidthHeight(self):
        glyph = self._glyph
        if glyph is None:
            return 0, 0
        # get the default layer width, that's what we use to draw
        # the background
        if glyph.layerSet is not None:
            glyph = glyph.layerSet.defaultLayer[glyph.name]
        bottom = self._descender
        top = max(self._capHeight, self._ascender, self._unitsPerEm + self._descender)
        width = glyph.width
        height = -bottom + top
        return width, height

    def fitScaleMetrics(self):
        """
        Scales and centers the viewport around the font’s metrics.
        """
        scrollArea = self._scrollArea
        if scrollArea:
            fitHeight = scrollArea.viewport().height()
        else:
            fitHeight = self.height()
        glyphWidth, glyphHeight = self._getGlyphWidthHeight()
        glyphHeight += self._noPointSizePadding * 2
        self.setScale(fitHeight / glyphHeight)
        self.centerOn(
            self.mapFromCanvas(
                QPointF(glyphWidth / 2, self._descender + self._unitsPerEm / 2)
            )
        )

    def fitScaleBBox(self):
        """
        Scales and centers the viewport around the *glyph*’s bounding box.
        """
        if self._glyph is None:
            return
        if self._glyph.bounds is None:
            self.fitScaleMetrics()
            return
        scrollArea = self._scrollArea
        if scrollArea:
            viewport = scrollArea.viewport()
            fitHeight = viewport.height()
            fitWidth = viewport.width()
        else:
            fitHeight = self.height()
            fitWidth = self.width()
        left, bottom, right, top = self._glyph.bounds
        glyphHeight = top - bottom
        glyphHeight += self._noPointSizePadding * 2
        glyphWidth = right - left
        glyphWidth += self._noPointSizePadding * 2
        self.setScale(min(fitHeight / glyphHeight, fitWidth / glyphWidth))
        self.centerOn(
            self.mapFromCanvas(
                QPointF(left + (right - left) / 2, bottom + (top - bottom) / 2)
            )
        )
        self.pointSizeModified.emit(self._impliedPointSize)

    def zoom(self, step, anchor="center"):
        """
        Zooms the view by *step* increments (with a scale factor of
        1.2^*step*), anchored to *anchor*:

        - QPoint_: center on that point
        - "cursor": center on the mouse cursor position
        - "center": center on the viewport
        - None: don’t anchor, i.e. stick to the viewport’s top-left.

        # TODO: improve docs from QGraphicsView descriptions.

        The default is "center".

        .. _QPoint: http://doc.qt.io/qt-5/qpoint.html
        """
        oldScale = self._scale
        newScale = self._scale * pow(1.2, step)
        scrollArea = self._scrollArea
        if newScale < 1e-2 or newScale > 1e3:
            return
        if scrollArea is not None:
            # compute new scrollbar position
            # http://stackoverflow.com/a/32269574/2037879
            hSB = scrollArea.horizontalScrollBar()
            vSB = scrollArea.verticalScrollBar()
            viewport = scrollArea.viewport()
            if isinstance(anchor, QPoint):
                pos = anchor
            elif anchor == "cursor":
                pos = self.mapFromGlobal(QCursor.pos())
            elif anchor == "center":
                pos = self.mapFromParent(
                    QPoint(viewport.width() / 2, viewport.height() / 2)
                )
            else:
                raise ValueError(f"invalid anchor value: {anchor}")
            scrollBarPos = QPointF(hSB.value(), vSB.value())
            deltaToPos = pos / oldScale
            delta = deltaToPos * (newScale - oldScale)
        self.setScale(newScale)
        self.update()
        if scrollArea is not None:
            hSB.setValue(scrollBarPos.x() + delta.x())
            vSB.setValue(scrollBarPos.y() + delta.y())

    # position mapping

    def mapFromCanvas(self, pos):
        """
        Maps *pos* from glyph canvas to this widget’s coordinates.

        Note that canvas coordinates are scale-independent while widget
        coordinates are not.
        """
        if self._drawingRect is None:
            self._calculateDrawingRect()
        xOffsetInv, yOffsetInv, _, _ = self._drawingRect
        x = (pos.x() - xOffsetInv) * self._scale
        y = (pos.y() - yOffsetInv) * (-self._scale) + self.height()
        return pos.__class__(x, y)

    def mapToCanvas(self, pos):
        """
        Maps *pos* from this widget’s to glyph canvas coordinates.

        Note that canvas coordinates are scale-independent while widget
        coordinates are not.
        """
        if self._drawingRect is None:
            self._calculateDrawingRect()
        xOffsetInv, yOffsetInv, _, _ = self._drawingRect
        x = pos.x() * self._inverseScale + xOffsetInv
        y = (pos.y() - self.height()) * (-self._inverseScale) + yOffsetInv
        return pos.__class__(x, y)

    def mapRectFromCanvas(self, rect):
        x, y, w, h = rect.getRect()
        origin = self.mapFromCanvas(QPointF(x, y))
        w *= self._scale
        h *= self._scale
        return rect.__class__(origin.x(), origin.y() - h, w, h)

    def mapRectToCanvas(self, rect):
        x, y, w, h = rect.getRect()
        origin = self.mapToCanvas(QPointF(x, y))
        w *= self._inverseScale
        h *= self._inverseScale
        return rect.__class__(origin.x(), origin.y() - h, w, h)

    # --------------------
    # Notification Support
    # --------------------

    def glyphChanged(self):
        # TODO: we could adjustSize() only when glyph width changes
        self.adjustSize()
        self.update()

    def fontChanged(self):
        self.setGlyph(self._glyph)

    # ---------------
    # Display Control
    # ---------------

    def drawingAttribute(self, attr, layerName):
        if layerName is None:
            return self._fallbackDrawingAttributes.get(attr)
        d = self._layerDrawingAttributes.get(layerName, {})
        return d.get(attr, attr == "showGlyphStroke" or None)

    def setDrawingAttribute(self, attr, value, layerName):
        if layerName is None:
            self._fallbackDrawingAttributes[attr] = value
        else:
            if layerName not in self._layerDrawingAttributes:
                self._layerDrawingAttributes[layerName] = {}
            self._layerDrawingAttributes[layerName][attr] = value
        self.update()

    def showFill(self):
        return self.drawingAttribute("showGlyphFill", None)

    def setShowFill(self, value):
        self.setDrawingAttribute("showGlyphFill", value, None)

    def showStroke(self):
        return self.drawingAttribute("showGlyphStroke", None)

    def setShowStroke(self, value):
        self.setDrawingAttribute("showGlyphStroke", value, None)

    def showMetrics(self):
        return self.drawingAttribute("showGlyphMargins", None)

    def setShowMetrics(self, value):
        self.setDrawingAttribute("showGlyphMargins", value, None)
        self.setDrawingAttribute("showFontVerticalMetrics", value, None)

    def showImage(self):
        return self.drawingAttribute("showGlyphImage", None)

    def setShowImage(self, value):
        self.setDrawingAttribute("showGlyphImage", value, None)

    def showMetricsTitles(self):
        return self.drawingAttribute("showFontVerticalMetricsTitles", None)

    def setShowMetricsTitles(self, value):
        self.setDrawingAttribute("showFontVerticalMetricsTitles", value, None)

    def showGuidelines(self):
        return self.drawingAttribute("showFontGuidelines", None)

    def setShowGuidelines(self, value):
        self.setDrawingAttribute("showFontGuidelines", value, None)
        self.setDrawingAttribute("showGlyphGuidelines", value, None)

    def showOnCurvePoints(self):
        return self.drawingAttribute("showGlyphOnCurvePoints", None)

    def setShowOnCurvePoints(self, value):
        self.setDrawingAttribute("showGlyphStartPoints", value, None)
        self.setDrawingAttribute("showGlyphOnCurvePoints", value, None)

    def showOffCurvePoints(self):
        return self.drawingAttribute("showGlyphOffCurvePoints", None)

    def setShowOffCurvePoints(self, value):
        self.setDrawingAttribute("showGlyphOffCurvePoints", value, None)

    def showPointCoordinates(self):
        return self.drawingAttribute("showGlyphPointCoordinates", None)

    def setShowPointCoordinates(self, value):
        self.setDrawingAttribute("showGlyphPointCoordinates", value, None)

    def showAnchors(self):
        return self.drawingAttribute("showGlyphAnchors", None)

    def setShowAnchors(self, value):
        self.setDrawingAttribute("showGlyphAnchors", value, None)

    def showBlues(self):
        return self.drawingAttribute("showFontPostscriptBlues", None)

    def setShowBlues(self, value):
        self.setDrawingAttribute("showFontPostscriptBlues", value, None)

    def showFamilyBlues(self):
        return self.drawingAttribute("showFontPostscriptFamilyBlues", None)

    def setShowFamilyBlues(self, value):
        self.setDrawingAttribute("showFontPostscriptFamilyBlues", value, None)

    def backgroundColor(self):
        return self._backgroundColor

    def setBackgroundColor(self, color):
        self._backgroundColor = color
        self.update()

    # ---------------
    # Drawing helpers
    # ---------------

    def drawBackground(self, painter):
        pass

    def drawGlyphLayer(self, painter, glyph, layerName, default=False):
        fontLayerName = None if default else layerName
        # draw the image
        if self.drawingAttribute("showGlyphImage", layerName):
            self.drawImage(painter, glyph, fontLayerName)
        # draw the blues
        if fontLayerName is None and self.drawingAttribute(
            "showFontPostscriptBlues", None
        ):
            self.drawBlues(painter, glyph, fontLayerName)
        if fontLayerName is None and self.drawingAttribute(
            "showFontPostscriptFamilyBlues", None
        ):
            self.drawFamilyBlues(painter, glyph, fontLayerName)
        # draw the metrics
        if fontLayerName is None and self.drawingAttribute("showGlyphMetrics", None):
            self.drawMetrics(painter, glyph, fontLayerName)

        layerName = None if glyph == self._glyph else layerName
        # draw the guidelines
        if (
            layerName is None
            and self.drawingAttribute("showFontGuidelines", None)
            or self.drawingAttribute("showGlyphGuidelines", None)
        ):
            self.drawGuidelines(painter, glyph, layerName)
        # draw the glyph
        if (
            self.drawingAttribute("showGlyphOnCurvePoints", layerName)
            or self.drawingAttribute("showGlyphOffCurvePoints", layerName)
            or self.drawingAttribute("showGlyphFill", layerName)
        ):
            self.drawFillAndPoints(painter, glyph, layerName)
        if self.drawingAttribute("showGlyphStroke", layerName):
            self.drawStroke(painter, glyph, layerName)
        if self.drawingAttribute("showGlyphAnchors", layerName):
            self.drawAnchors(painter, glyph, layerName)

    def drawImage(self, painter, glyph, layerName):
        drawing.drawGlyphImage(painter, glyph, self._inverseScale)

    def drawBlues(self, painter, glyph, layerName):
        drawing.drawFontPostscriptBlues(painter, glyph, self._inverseScale)

    def drawFamilyBlues(self, painter, glyph, layerName):
        drawing.drawFontPostscriptFamilyBlues(painter, glyph, self._inverseScale)

    def drawMetrics(self, painter, glyph, layerName):
        drawVMetrics = layerName is None and self.drawingAttribute(
            "showFontVerticalMetrics", None
        )
        drawText = (
            self.drawingAttribute("showFontVerticalMetricsTitles", layerName)
            and self._impliedPointSize > GlyphViewMinSizeForDetails
        )
        drawing.drawGlyphMetrics(
            painter,
            glyph,
            self._inverseScale,
            drawVMetrics=drawVMetrics,
            drawText=drawText,
        )

    def drawGuidelines(self, painter, glyph, layerName):
        drawText = self._impliedPointSize > GlyphViewMinSizeForDetails
        if self.drawingAttribute("showFontGuidelines", layerName):
            drawing.drawFontGuidelines(
                painter, glyph, self._inverseScale, self._drawingRect, drawText=drawText
            )
        if self.drawingAttribute("showGlyphGuidelines", layerName):
            drawing.drawGlyphGuidelines(
                painter, glyph, self._inverseScale, self._drawingRect, drawText=drawText
            )

    def drawFillAndPoints(self, painter, glyph, layerName):
        drawFill = self.drawingAttribute("showGlyphFill", layerName)
        drawing.drawGlyphFillAndStroke(
            painter, glyph, self._inverseScale, drawFill=drawFill, drawStroke=False
        )
        if not self._impliedPointSize > GlyphViewMinSizeForDetails:
            return
        drawStartPoints = self.drawingAttribute("showGlyphStartPoints", layerName)
        drawOnCurves = self.drawingAttribute("showGlyphOnCurvePoints", layerName)
        drawOffCurves = self.drawingAttribute("showGlyphOffCurvePoints", layerName)
        drawCoordinates = self.drawingAttribute("showGlyphPointCoordinates", layerName)
        drawing.drawGlyphPoints(
            painter,
            glyph,
            self._inverseScale,
            drawStartPoints=drawStartPoints,
            drawOnCurves=drawOnCurves,
            drawOffCurves=drawOffCurves,
            drawCoordinates=drawCoordinates,
            backgroundColor=self._backgroundColor,
        )

    def drawStroke(self, painter, glyph, layerName):
        showStroke = self.drawingAttribute("showGlyphStroke", layerName)
        drawing.drawGlyphFillAndStroke(
            painter,
            glyph,
            self._inverseScale,
            drawFill=False,
            drawComponentFill=False,
            drawStroke=showStroke,
        )

    def drawAnchors(self, painter, glyph, layerName):
        if not self._impliedPointSize > GlyphViewMinSizeForDetails:
            return
        drawing.drawGlyphAnchors(painter, glyph, self._inverseScale, self._drawingRect)

    def drawForeground(self, painter):
        pass

    # ---------------
    # QWidget methods
    # ---------------

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setFont(UIFont)
        painter.setRenderHint(QPainter.Antialiasing)
        rect = event.rect()

        # draw the background
        painter.fillRect(rect, self._backgroundColor)
        if self._glyph is None:
            return

        # apply the overall scale
        painter.save()
        # + translate and flip
        painter.translate(0, self.height())
        painter.scale(self._scale, -self._scale)

        # move into position
        xOffsetInv, yOffsetInv, _, _ = self._drawingRect
        painter.translate(-xOffsetInv, -yOffsetInv)

        # gather the layers
        layerSet = self._glyph.layerSet
        if layerSet is None:
            layers = [(self._glyph, None, True)]
        else:
            glyphName = self._glyph.name
            layers = []
            for layerName in reversed(layerSet.layerOrder):
                layer = layerSet[layerName]
                if glyphName not in layer:
                    continue
                glyph = layer[glyphName]
                layers.append((glyph, layerName, layer == layerSet.defaultLayer))

        self.drawBackground(painter)
        for glyph, layerName, default in layers:
            self.drawGlyphLayer(painter, glyph, layerName, default)
        self.drawForeground(painter)
        painter.restore()

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self._calculateDrawingRect()

    def sizeHint(self):
        # pick the width and height
        glyphWidth, glyphHeight = self._getGlyphWidthHeight()
        glyphWidth = glyphWidth * self._scale
        glyphHeight = glyphHeight * self._scale
        xOffset = 1000 * 2 * self._scale
        yOffset = xOffset
        width = glyphWidth + xOffset
        height = glyphHeight + yOffset
        # calculate and store the vertical centering offset
        scrollArea = self._scrollArea
        if scrollArea:
            maxHeight = max(height, scrollArea.viewport().height())
        else:
            maxHeight = height
        self._verticalCenterYBuffer = (maxHeight - glyphHeight) / 2.0
        return QSize(width, height)

    def showEvent(self, event):
        super().showEvent(event)
        if hasattr(self, "_fitViewport"):
            self._calculateDrawingRect()
            self.fitScaleBBox()
            del self._fitViewport

    def wheelEvent(self, event):
        if event.modifiers() & platformSpecific.scaleModifier():
            step = event.angleDelta().y() / 120.0
            self.zoom(step, event.pos())
            self.pointSizeModified.emit(self._impliedPointSize)
            event.accept()
        else:
            super().wheelEvent(event)


class GlyphView(QScrollArea):
    glyphWidgetClass = GlyphWidget

    def __init__(self, parent=None):
        super().__init__(parent)
        self.grabGesture(Qt.PinchGesture)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.setWidgetResizable(True)

        self._glyphWidget = self.glyphWidgetClass(self)
        self._glyphWidget.setScrollArea(self)

        self.pointSizeModified = self._glyphWidget.pointSizeModified

    # -------------
    # Notifications
    # -------------

    def _subscribeToGlyph(self, glyph):
        if glyph is not None:
            glyph.addObserver(self, "_glyphChanged", "Glyph.Changed")
            layerSet = glyph.layerSet
            if layerSet is not None:
                layerSet.addObserver(self, "_glyphChanged", "LayerSet.LayerChanged")
            font = glyph.font
            if font is not None:
                font.info.addObserver(self, "_fontChanged", "Info.Changed")

    def _unsubscribeFromGlyph(self):
        if self._glyphWidget is not None:
            glyph = self._glyphWidget.glyph()
            if glyph is not None:
                glyph.removeObserver(self, "Glyph.Changed")
                layerSet = glyph.layerSet
                if layerSet is not None:
                    layerSet.removeObserver(self, "LayerSet.LayerChanged")
                font = glyph.font
                if font is not None:
                    font.info.removeObserver(self, "Info.Changed")

    def _glyphChanged(self, notification):
        self._glyphWidget.glyphChanged()

    def _fontChanged(self, notification):
        self._glyphWidget.fontChanged()

    # --------------
    # Public Methods
    # --------------

    def pointSize(self):
        return self._glyphWidget.pointSize()

    def setPointSize(self, pointSize):
        self._glyphWidget.setPointSize(pointSize)

    def scale(self):
        return self._glyphWidget.scale()

    def setScale(self, scale):
        self._glyphWidget.setScale(scale)

    def glyph(self):
        return self._glyphWidget.glyph()

    def setGlyph(self, glyph):
        self._unsubscribeFromGlyph()
        self._subscribeToGlyph(glyph)
        self._glyphWidget.setGlyph(glyph)

    def drawingAttribute(self, attr, layerName=None):
        return self._glyphWidget.drawingAttribute(attr, layerName)

    def setDrawingAttribute(self, attr, value, layerName=None):
        self._glyphWidget.setDrawingAttribute(attr, value, layerName)

    def fitScaleBBox(self):
        self._glyphWidget.fitScaleBBox()

    def zoom(self, factor, anchor="center"):
        self._glyphWidget.zoom(factor, anchor=anchor)

    # convenience

    def showFill(self):
        return self.drawingAttribute("showGlyphFill")

    def setShowFill(self, value):
        self.setDrawingAttribute("showGlyphFill", value)

    def showStroke(self):
        return self.drawingAttribute("showGlyphStroke")

    def setShowStroke(self, value):
        self.setDrawingAttribute("showGlyphStroke", value)

    def showMetrics(self):
        return self.drawingAttribute("showGlyphMargins")

    def setShowMetrics(self, value):
        self.setDrawingAttribute("showGlyphMargins", value)
        self.setDrawingAttribute("showFontVerticalMetrics", value)
        self.setDrawingAttribute("showFontVerticalMetricsTitles", value)

    def showImage(self):
        return self.drawingAttribute("showGlyphImage")

    def setShowImage(self, value):
        self.setDrawingAttribute("showGlyphImage", value)

    def showOnCurvePoints(self):
        return self.drawingAttribute("showGlyphOnCurvePoints")

    def setShowOnCurvePoints(self, value):
        self.setDrawingAttribute("showGlyphOnCurvePoints", value)

    def showOffCurvePoints(self):
        return self.drawingAttribute("showGlyphOffCurvePoints")

    def setShowOffCurvePoints(self, value):
        self.setDrawingAttribute("showGlyphOffCurvePoints", value)

    def showPointCoordinates(self):
        return self.drawingAttribute("showGlyphPointCoordinates")

    def setShowPointCoordinates(self, value):
        self.setDrawingAttribute("showGlyphPointCoordinates", value)

    def showAnchors(self):
        return self.drawingAttribute("showGlyphAnchors")

    def setShowAnchors(self, value):
        self.setDrawingAttribute("showGlyphAnchors", value)

    def showBlues(self):
        return self.drawingAttribute("showFontPostscriptBlues")

    def setShowBlues(self, value):
        self.setDrawingAttribute("showFontPostscriptBlues", value)

    def showFamilyBlues(self):
        return self.drawingAttribute("showFontPostscriptFamilyBlues")

    def setShowFamilyBlues(self, value):
        self.setDrawingAttribute("showFontPostscriptFamilyBlues", value)

    def backgroundColor(self):
        return self._glyphWidget.backgroundColor()

    def setBackgroundColor(self, color):
        self._glyphWidget.setBackgroundColor(color)

    # ----------
    # Qt methods
    # ----------

    def event(self, event):
        if event.type() == QEvent.Gesture:
            return self.gestureEvent(event)
        return super().event(event)

    def gestureEvent(self, event):
        gesture = event.gesture(Qt.PinchGesture)
        if gesture:
            self.pinchTriggered(gesture)
            return True
        return False

    def pinchTriggered(self, gesture):
        changeFlags = gesture.changeFlags()
        if changeFlags & QPinchGesture.ScaleFactorChanged:
            self.zoom(gesture.scaleFactor() - gesture.lastScaleFactor(), "cursor")



"""
The *qPainterPathFactory* submodule
-----------------------------

The *qPainterPathFactory* submodule provides a QPainterPath_ representation of
a Glyph_’s outline.

QPainterPath_ is the Qt class for Bézier paths. It accomodates cubic or
quadratic outlines and has a pen interface.

You can then draw such paths on screen with the QPainter_ method
``drawPath()``.

.. _Glyph: http://ts-defcon.readthedocs.org/en/ufo3/objects/glyph.html
.. _QPainter: http://doc.qt.io/qt-5/qpainter.html
.. _QPainterPath: http://doc.qt.io/qt-5/qpainterpath.html
"""

from fontTools.pens.qtPen import QtPen
from PyQt5.QtCore import Qt


def QPainterPathFactory(glyph):
    pen = QtPen(glyph.layer)
    glyph.draw(pen)
    pen.path.setFillRule(Qt.WindingFill)
    return pen.path



import os
import sys

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QFont, QKeySequence

# -----------
# File dialog
# -----------


def treatPackageAsFile():
    return sys.platform == "darwin"


# -----
# Fonts
# -----


def fontSizeDelta():
    return int(sys.platform == "darwin")


def UIFontOverride():
    if sys.platform == "win32":
        font = QFont()
        font.setFamily("Segoe UI")
        font.setPointSize(9)
        return font
    return None


# -------------
# Key sequences
# -------------


def altDeleteSequence():
    if sys.platform == "darwin":
        return "Backspace"
    return None


def altRedoSequence():
    if sys.platform == "win32":
        return "Ctrl+Shift+Z"
    return None


def closeKeySequence():
    if sys.platform == "win32":
        return "Ctrl+W"
    return QKeySequence.Close


def previousTabSequence():
    if sys.platform == "win32":
        return "Ctrl+Shift+Tab"
    return QKeySequence.PreviousChild


def combinedModifiers():
    # on Windows, Ctrl+Alt is reserved by the system. use WinKey+Alt
    if sys.platform == "win32":
        return Qt.MetaModifier | Qt.AltModifier
    return Qt.ControlModifier | Qt.AltModifier


def isDeleteEvent(event):
    if event.matches(QKeySequence.Delete):
        return True
    if sys.platform == "darwin" and event.key() == Qt.Key_Backspace:
        return True
    modifiers = event.modifiers()
    if modifiers & Qt.ShiftModifier or modifiers & Qt.AltModifier:
        modifiers_ = modifiers & ~Qt.ShiftModifier & ~Qt.AltModifier
        event_ = event.__class__(
            event.type(),
            event.key(),
            modifiers_,
            event.text(),
            event.isAutoRepeat(),
            event.count(),
        )
        return event_.matches(QKeySequence.Delete)
    return False


# -------
# Margins
# -------


def needsTighterMargins():
    return sys.platform == "darwin"


def widen():
    return sys.platform == "win32"


# --------
# Menu bar
# --------


def useGlobalMenuBar():
    if sys.platform == "darwin":
        return True
    elif sys.platform.startswith("linux"):
        env = os.environ
        if (
            env.get("XDG_CURRENT_DESKTOP") == "Unity"
            and len(env.get("UBUNTU_MENUPROXY", "")) > 1
        ):
            return True
    return False


def mergeOpenAndImport():
    return sys.platform == "darwin"


def windowCommandsInMenu():
    return sys.platform == "darwin"


def setAppName():
    if sys.platform == "darwin":
        try:
            from Foundation import NSBundle

            bundle = NSBundle.mainBundle()
            if bundle:
                app_info = bundle.localizedInfoDictionary() or bundle.infoDictionary()
                if app_info:
                    app_info["CFBundleName"] = "TruFont"
        except ImportError as e:
            print(f"Could not set title: {e}")
            pass


# -----------
# Main window
# -----------


def appNameInTitle():
    if sys.platform == "darwin":
        return False
    return True


def shouldSpawnDocument():
    return sys.platform != "darwin"


# -----------
# Message box
# -----------


def showAppIconInDialog():
    return sys.platform == "darwin"


# -----------
# Rubber band
# -----------


def useBuiltinRubberBand():
    return sys.platform == "darwin"


# ----------
# Stylesheet
# ----------


def appStyleSheet():
    if sys.platform == "win32":
        return "QStatusBar::item { border: none; }"
    elif sys.platform == "darwin":
        return "QToolTip { background-color: white; }"
    return None



import os
import pickle
from collections import OrderedDict

from fontTools.feaLib.error import FeatureLibError
from fontTools.pens.svgPathPen import SVGPathPen
from fontTools.svgLib import SVGPath
from fontTools.ufoLib.glifLib import readGlyphFromString
from PyQt5.QtCore import QEvent, QMimeData, QObject, QSize, QStandardPaths, Qt
from PyQt5.QtGui import QColor, QKeySequence, QPainter, QPainterPath
from PyQt5.QtWidgets import (
    QApplication,
    QFileDialog,
    QHBoxLayout,
    QMessageBox,
    QShortcut,
    QStackedWidget,
    QVBoxLayout,
    QWidget,
)

from defconQt.controls.glyphCellView import GlyphCellView
from defconQt.windows.baseWindows import BaseWindow
from trufont.controls.exportDialog import ExportDialog
from trufont.controls.fileMessageBoxes import CloseMessageBox, ReloadMessageBox
from trufont.controls.fontDialogs import AddGlyphsDialog, SortDialog
from trufont.controls.glyphCanvasView import GlyphCanvasView
from trufont.controls.glyphDialogs import FindDialog
from trufont.controls.propertiesView import PropertiesView
from trufont.controls.statusBar import StatusBar
from trufont.controls.tabWidget import TabWidget
from trufont.controls.toolBar import ToolBar
from trufont.objects import settings
from trufont.objects.menu import Entries
from trufont.tools import errorReports, platformSpecific
from trufont.tools.uiMethods import deleteUISelection, removeUIGlyphElements
from trufont.windows.fontFeaturesWindow import FontFeaturesWindow
from trufont.windows.fontInfoWindow import FontInfoWindow
from trufont.windows.groupsWindow import GroupsWindow
from trufont.windows.kerningWindow import KerningWindow
from trufont.windows.metricsWindow import MetricsWindow

_path = QPainterPath()
_path.moveTo(5, 8)
_path.lineTo(23, 8)
_path.lineTo(23, 10)
_path.lineTo(5, 10)
_path.closeSubpath()
_path.moveTo(5, 13)
_path.lineTo(23, 13)
_path.lineTo(23, 15)
_path.lineTo(5, 15)
_path.closeSubpath()
_path.moveTo(5, 18)
_path.lineTo(23, 18)
_path.lineTo(23, 20)
_path.lineTo(5, 20)
_path.closeSubpath()


def _textForGlyphs(glyphs):
    names = []
    for glyph in glyphs:
        char = ""
        if glyph.unicode and chr(glyph.unicode) != glyph.name:
            char = " ({})".format(chr(glyph.unicode))
        names.append(f"/{glyph.name}{char}")
    return " ".join(names)


class PageWidget(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(1)

    def addWidget(self, widget):
        self.layout().addWidget(widget)

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.fillRect(event.rect(), QColor(235, 235, 235))


class PreviewEventFilter(QObject):
    filterKeyEvents = (QEvent.ShortcutOverride, QEvent.KeyPress, QEvent.KeyRelease)

    def eventFilter(self, obj, event):
        if not obj.isWidgetType():
            return False
        # or should we reset on WindowActivate?
        if event.type() == QEvent.WindowDeactivate:
            self.parent()._setGlyphPreview(False)
        if event.type() in self.filterKeyEvents:
            if not event.isAutoRepeat() and event.key() == Qt.Key_Space:
                self.parent()._setGlyphPreview(event.type() != QEvent.KeyRelease)
                event.accept()
                return True
        return False


class FontWindow(BaseWindow):
    def __init__(self, font, parent=None):
        super().__init__(parent)
        self._font = None

        self._infoWindow = None
        self._featuresWindow = None
        self._groupsWindow = None
        self._kerningWindow = None
        self._metricsWindow = None

        self.toolBar = ToolBar(self)
        self.toolBar.setTools(t() for t in QApplication.instance().drawingTools())

        self.glyphCellView = GlyphCellView(self)
        self.glyphCellView.glyphActivated.connect(self.openGlyphTab)
        self.glyphCellView.glyphsDropped.connect(self._orderChanged)
        self.glyphCellView.selectionChanged.connect(self._selectionChanged)
        self.glyphCellView.setAcceptDrops(True)
        self.glyphCellView.setCellRepresentationName("TruFont.GlyphCell")
        self.glyphCellView.setFrameShape(self.glyphCellView.NoFrame)
        self.glyphCellView.setFocus()

        self.tabWidget = TabWidget(self)
        self.tabWidget.setAutoHide(True)
        self.tabWidget.setHeroFirstTab(True)
        self.tabWidget.addTab(self.tr("Font"))

        self.stackWidget = QStackedWidget(self)
        self.stackWidget.addWidget(self.glyphCellView)
        self.tabWidget.currentTabChanged.connect(self._tabChanged)
        self.tabWidget.tabRemoved.connect(
            lambda index: self.stackWidget.removeWidget(self.stackWidget.widget(index))
        )
        self.stackWidget.currentChanged.connect(self._widgetChanged)

        self.propertiesView = PropertiesView(font, self)
        self.propertiesView.hide()

        self.statusBar = StatusBar(self)
        self.statusBar.setMinimumSize(32)
        self.statusBar.setMaximumSize(128)
        self.statusBar.sizeChanged.connect(self._sizeChanged)

        self.setFont_(font)

        app = QApplication.instance()
        app.dispatcher.addObserver(
            self, "_drawingToolRegistered", "drawingToolRegistered"
        )
        app.dispatcher.addObserver(
            self, "_drawingToolUnregistered", "drawingToolUnregistered"
        )
        app.dispatcher.addObserver(
            self, "_glyphViewGlyphsChanged", "glyphViewGlyphsChanged"
        )

        layout = QHBoxLayout(self)
        layout.addWidget(self.toolBar)
        vLayout = QVBoxLayout()
        vLayout.addWidget(self.tabWidget)
        pageWidget = PageWidget()
        pageWidget.addWidget(self.stackWidget)
        pageWidget.addWidget(self.statusBar)
        vLayout.addWidget(pageWidget)
        layout.addLayout(vLayout)
        layout.addWidget(self.propertiesView)
        layout.setContentsMargins(0, 2, 0, 0)
        layout.setSpacing(2)

        elements = [
            ("Ctrl+D", self.deselect),
            (platformSpecific.closeKeySequence(), self.closeGlyphTab),
            # XXX: does this really not warrant widget focus?
            (QKeySequence.Delete, self.delete),
            ("Shift+" + QKeySequence(QKeySequence.Delete).toString(), self.delete),
            ("Z", lambda: self.zoom(1)),
            ("X", lambda: self.zoom(-1)),
        ]
        e = platformSpecific.altDeleteSequence()
        if e is not None:
            elements.append((e, self.delete))
        e = platformSpecific.altRedoSequence()
        if e is not None:
            elements.append((e, self.redo))
        for keys, callback in elements:
            shortcut = QShortcut(QKeySequence(keys), self)
            shortcut.activated.connect(callback)

        self.installEventFilter(PreviewEventFilter(self))

        self.readSettings()
        self.propertiesView.activeLayerModified.connect(self._activeLayerModified)
        self.statusBar.sizeChanged.connect(self.writeSettings)

    def readSettings(self):
        geometry = settings.fontWindowGeometry()
        if geometry:
            self.restoreGeometry(geometry)
        cellSize = settings.glyphCellSize()
        self.statusBar.setSize(cellSize)
        hidden = settings.propertiesHidden()
        if not hidden:
            self.properties()

    def writeSettings(self):
        settings.setFontWindowGeometry(self.saveGeometry())
        settings.setGlyphCellSize(self.glyphCellView.cellSize()[0])
        settings.setPropertiesHidden(self.propertiesView.isHidden())

    def menuBar(self):
        return self.layout().menuBar()

    def setMenuBar(self, menuBar):
        self.layout().setMenuBar(menuBar)

    def setupMenu(self, menuBar):
        app = QApplication.instance()

        fileMenu = menuBar.fetchMenu(Entries.File)
        fileMenu.fetchAction(Entries.File_New)
        fileMenu.fetchAction(Entries.File_Open)
        fileMenu.fetchMenu(Entries.File_Open_Recent)
        if not platformSpecific.mergeOpenAndImport():
            fileMenu.fetchAction(Entries.File_Import)
        fileMenu.addSeparator()
        fileMenu.fetchAction(Entries.File_Save, self.saveFile)
        fileMenu.fetchAction(Entries.File_Save_As, self.saveFileAs)
        fileMenu.fetchAction(Entries.File_Save_All)
        fileMenu.fetchAction(Entries.File_Reload, self.reloadFile)
        fileMenu.addSeparator()
        fileMenu.fetchAction(Entries.File_Export, self.exportFile)
        fileMenu.fetchAction(Entries.File_Exit)

        editMenu = menuBar.fetchMenu(Entries.Edit)
        self._undoAction = editMenu.fetchAction(Entries.Edit_Undo, self.undo)
        self._redoAction = editMenu.fetchAction(Entries.Edit_Redo, self.redo)
        editMenu.addSeparator()
        cut = editMenu.fetchAction(Entries.Edit_Cut, self.cut)
        copy = editMenu.fetchAction(Entries.Edit_Copy, self.copy)
        copyComponent = editMenu.fetchAction(
            Entries.Edit_Copy_As_Component, self.copyAsComponent
        )
        paste = editMenu.fetchAction(Entries.Edit_Paste, self.paste)
        self._clipboardActions = (cut, copy, copyComponent, paste)
        editMenu.fetchAction(Entries.Edit_Select_All, self.selectAll)
        # editMenu.fetchAction(Entries.Edit_Deselect, self.deselect)
        editMenu.fetchAction(Entries.Edit_Find, self.findGlyph)
        editMenu.addSeparator()
        editMenu.fetchAction(Entries.Edit_Settings)

        viewMenu = menuBar.fetchMenu(Entries.View)
        viewMenu.fetchAction(Entries.View_Zoom_In, lambda: self.zoom(1))
        viewMenu.fetchAction(Entries.View_Zoom_Out, lambda: self.zoom(-1))
        viewMenu.fetchAction(Entries.View_Reset_Zoom, self.resetZoom)
        viewMenu.addSeparator()
        viewMenu.fetchAction(Entries.View_Next_Tab, lambda: self.tabOffset(1))
        viewMenu.fetchAction(Entries.View_Previous_Tab, lambda: self.tabOffset(-1))
        viewMenu.fetchAction(Entries.View_Next_Glyph, lambda: self.glyphOffset(1))
        viewMenu.fetchAction(Entries.View_Previous_Glyph, lambda: self.glyphOffset(-1))
        viewMenu.fetchAction(Entries.View_Layer_Up, lambda: self.layerOffset(-1))
        viewMenu.fetchAction(Entries.View_Layer_Down, lambda: self.layerOffset(1))
        viewMenu.addSeparator()
        viewMenu.fetchAction(Entries.View_Show_Points)
        coordinatesSubmenu = viewMenu.fetchMenu(Entries.View_Show_Coordinates)
        coordinatesSubmenu.fetchAction(Entries.View_Show_Coordinates_When_Selected)
        coordinatesSubmenu.fetchAction(Entries.View_Show_Point_Coordinates)
        coordinatesSubmenu.fetchAction(Entries.View_Show_Bezier_Handles_Coordinates)
        viewMenu.fetchAction(Entries.View_Show_Metrics)
        viewMenu.fetchAction(Entries.View_Show_Images)
        viewMenu.fetchAction(Entries.View_Show_Guidelines)

        fontMenu = menuBar.fetchMenu(Entries.Font)
        fontMenu.fetchAction(Entries.Font_Font_Info, self.fontInfo)
        fontMenu.fetchAction(Entries.Font_Font_Features, self.fontFeatures)
        fontMenu.addSeparator()
        fontMenu.fetchAction(Entries.Font_Add_Glyphs, self.addGlyphs)
        fontMenu.fetchAction(Entries.Font_Sort, self.sortGlyphs)

        # glyphMenu = menuBar.fetchMenu(self.tr("&Glyph"))
        # self._layerAction = glyphMenu.fetchAction(
        #     self.tr("&Layer Actions…"), self.layerActions, "L")

        menuBar.fetchMenu(Entries.Scripts)

        windowMenu = menuBar.fetchMenu(Entries.Window)
        windowMenu.fetchAction(Entries.Window_Groups, self.groups)
        windowMenu.fetchAction(Entries.Window_Kerning, self.kerning)
        windowMenu.fetchAction(Entries.Window_Metrics, self.metrics)
        windowMenu.fetchAction(Entries.Window_Scripting)
        windowMenu.fetchAction(Entries.Window_Properties, self.properties)
        windowMenu.addSeparator()
        action = windowMenu.fetchAction(Entries.Window_Output)
        action.setEnabled(app.outputWindow is not None)

        helpMenu = menuBar.fetchMenu(Entries.Help)
        helpMenu.fetchAction(Entries.Help_Documentation)
        helpMenu.fetchAction(Entries.Help_Report_An_Issue)
        helpMenu.addSeparator()
        helpMenu.fetchAction(Entries.Help_About)

        self._updateGlyphActions()

    # --------------
    # Custom methods
    # --------------

    def font_(self):
        return self._font

    def setFont_(self, font):
        if self._font is not None:
            self._font.removeObserver(self, "Font.Changed")
            self._font.removeObserver(self, "Font.GlyphOrderChanged")
            self._font.removeObserver(self, "Font.SortDescriptorChanged")
        self._font = font
        self.setWindowTitle(self.fontTitle())
        if font is None:
            return
        self._updateGlyphsFromGlyphOrder()
        font.addObserver(self, "_fontChanged", "Font.Changed")
        font.addObserver(self, "_glyphOrderChanged", "Font.GlyphOrderChanged")
        font.addObserver(self, "_sortDescriptorChanged", "Font.SortDescriptorChanged")

    def fontTitle(self):
        if self._font is None:
            return None
        path = self._font.path or self._font.binaryPath
        if path is not None:
            return os.path.basename(path.rstrip(os.sep))
        return self.tr("Untitled")

    def isGlyphTab(self):
        return bool(self.stackWidget.currentIndex())

    def openGlyphTab(self, glyph):
        # if a tab with this glyph exists already, switch to it
        for index in range(self.stackWidget.count()):
            if not index:
                continue
            view = self.stackWidget.widget(index)
            if list(view.glyphs()) == [glyph]:
                self.tabWidget.setCurrentTab(index)
                return
        # spawn
        widget = GlyphCanvasView(self)
        widget.setInputNames([glyph.name])
        widget.activeGlyphChanged.connect(self._selectionChanged)
        widget.glyphNamesChanged.connect(self._namesChanged)
        widget.pointSizeModified.connect(self.statusBar.setSize)
        widget.toolModified.connect(self.toolBar.setCurrentTool)
        # add
        self.tabWidget.addTab(_textForGlyphs([glyph]))
        self.stackWidget.addWidget(widget)
        # activate
        self.tabWidget.setCurrentTab(-1)

    def closeGlyphTab(self):
        index = self.stackWidget.currentIndex()
        if index:
            self.tabWidget.removeTab(index)

    def maybeSaveBeforeExit(self):
        if self._font.dirty:
            ret = CloseMessageBox.getCloseDocument(self, self.fontTitle())
            if ret == QMessageBox.Save:
                self.saveFile()
                return True
            elif ret == QMessageBox.Discard:
                return True
            return False
        return True

    # -------------
    # Notifications
    # -------------

    # app

    def _drawingToolRegistered(self, notification):
        toolClass = notification.data["tool"]
        index = self.stackWidget.currentIndex()
        parent = self.stackWidget.currentWidget() if index else None
        self.toolBar.addTool(toolClass(parent=parent))

    def _drawingToolUnregistered(self, notification):
        toolClass = notification.data["tool"]
        for tool in self.toolBar.tools():
            if isinstance(tool, toolClass):
                self.toolBar.removeTool(tool)
                return
        raise ValueError(f"couldn't find tool to unregister: {toolClass}")

    def _glyphViewGlyphsChanged(self, notification):
        self._updateGlyphActions()

    # widgets

    def _activeLayerModified(self):
        if self.isGlyphTab():
            widget = self.stackWidget.currentWidget()
            index = self.sender().currentIndex().row()
            layers = self._font.layers
            layer = layers[layers.layerOrder[index]]
            currentGlyph = widget.activeGlyph()
            # XXX: adjust TLayer.get and use it
            if currentGlyph.name in layer:
                glyph = layer[currentGlyph.name]
            else:
                glyph = layer.newGlyph(currentGlyph.name)
            widget.setActiveGlyph(glyph)

    def _namesChanged(self):
        sender = self.sender()
        index = self.stackWidget.indexOf(sender)
        self.tabWidget.setTabName(index, _textForGlyphs(sender.glyphs()))

    def _sizeChanged(self):
        size = self.statusBar.size()
        if self.isGlyphTab():
            widget = self.stackWidget.currentWidget()
            widget.setPointSize(size)
        else:
            self.glyphCellView.setCellSize(size)

    def _tabChanged(self, index):
        self.statusBar.setShouldPropagateSize(not index)
        # we need to hide, then setParent, then show
        self.stackWidget.currentWidget().hide()
        newWidget = self.stackWidget.widget(index)
        if index:
            for tool in self.toolBar.tools():
                tool.setParent(newWidget)
        self.stackWidget.setCurrentIndex(index)
        newWidget.setFocus(Qt.OtherFocusReason)

    def _toolChanged(self, tool):
        widget = self.stackWidget.currentWidget()
        ok = widget.setCurrentTool(tool)
        # the glyph view NAKed the change (in mouseDown)
        # set back the current tool in the toolbar
        if not ok:
            self.toolBar.setCurrentTool(widget.currentTool())

    def _widgetChanged(self, index):
        # update current glyph
        self._updateCurrentGlyph()
        # update undo/redo
        self._updateGlyphActions()
        # update slider
        if self.isGlyphTab():
            lo, hi, unit = 0, 900_000, " pt"
            widget = self.stackWidget.currentWidget()
            size = widget.pointSize()
        else:
            lo, hi, unit = 32, 128, None
            size = self.glyphCellView.cellSize()[0]
        self.statusBar.setMinimumSize(lo)
        self.statusBar.setMaximumSize(hi)
        self.statusBar.setSize(size)
        self.statusBar.setUnit(unit)
        self.statusBar.setTextVisible(not self.isGlyphTab())
        # update and connect setCurrentTool
        try:
            self.toolBar.currentToolChanged.disconnect()
        except TypeError:
            pass
        if not index:
            return
        widget = self.stackWidget.currentWidget()
        widget.setCurrentTool(self.toolBar.currentTool())
        self.toolBar.currentToolChanged.connect(self._toolChanged)

    def _orderChanged(self):
        # TODO: reimplement when we start showing glyph subsets
        glyphs = self.glyphCellView.glyphs()
        self._font.glyphOrder = [glyph.name for glyph in glyphs]

    def _selectionChanged(self):
        if self.isGlyphTab():
            activeGlyph = self.stackWidget.currentWidget().activeGlyph()
        else:
            activeGlyph = self.glyphCellView.lastSelectedGlyph()
            # selection text
            # TODO: this should probably be internal to the label
            selection = self.glyphCellView.selection()
            if selection is not None:
                count = len(selection)
                if count == 1:
                    glyph = self.glyphCellView.glyphsForIndexes(selection)[0]
                    text = "%s " % glyph.name
                else:
                    text = ""
                if count:
                    text = self.tr(f"{text}(%n selected)", n=count)
            else:
                text = ""
            self.statusBar.setText(text)
        # currentGlyph
        app = QApplication.instance()
        app.setCurrentGlyph(activeGlyph)
        # actions
        self._updateGlyphActions()

    # defcon

    def _fontChanged(self, notification):
        font = notification.object
        self.setWindowModified(font.dirty)

    def _glyphOrderChanged(self, notification):
        self._updateGlyphsFromGlyphOrder()

    def _updateGlyphsFromGlyphOrder(self):
        font = self._font
        glyphOrder = font.glyphOrder
        if glyphOrder:
            glyphCount = 0
            glyphs = []
            for glyphName in glyphOrder:
                if glyphName in font:
                    glyph = font[glyphName]
                    glyphCount += 1
                else:
                    glyph = font.get(glyphName, asTemplate=True)
                glyphs.append(glyph)
            if glyphCount < len(font):
                # if some glyphs in the font are not present in the glyph
                # order, loop again to add them at the end
                for glyph in font:
                    if glyph not in glyphs:
                        glyphs.append(glyph)
                font.disableNotifications(observer=self)
                font.glyphOrder = [glyph.name for glyph in glyphs]
                font.enableNotifications(observer=self)
        else:
            glyphs = list(font)
            font.disableNotifications(observer=self)
            font.glyphOrder = [glyph.name for glyph in glyphs]
            font.enableNotifications(observer=self)
        self.glyphCellView.setGlyphs(glyphs)

    def _sortDescriptorChanged(self, notification):
        font = notification.object
        descriptors = notification.data["newValue"]
        if descriptors is None:
            return
        if descriptors[0]["type"] == "glyphSet":
            glyphNames = descriptors[0]["glyphs"]
        else:
            glyphNames = font.unicodeData.sortGlyphNames(font.keys(), descriptors)
        font.glyphOrder = glyphNames

    # ------------
    # Menu methods
    # ------------

    # File

    def saveFile(self, path=None, ufoFormatVersion=3):
        if path is None and self._font.path is None:
            self.saveFileAs()
        else:
            if path is None:
                path = self._font.path
            self._font.save(path, ufoFormatVersion)

    def saveFileAs(self):
        fileFormats = OrderedDict(
            [
                (self.tr("UFO Font version 3 {}").format("(*.ufo)"), 3),
                (self.tr("UFO Font version 2 {}").format("(*.ufo)"), 2),
            ]
        )
        state = settings.saveFileDialogState()
        path = self._font.path or self._font.binaryPath
        if path:
            directory = os.path.dirname(path)
        else:
            directory = (
                None
                if state
                else QStandardPaths.standardLocations(QStandardPaths.DocumentsLocation)[
                    0
                ]
            )
        # TODO: switch to directory dlg on platforms that need it
        dialog = QFileDialog(
            self, self.tr("Save File"), directory, ";;".join(fileFormats.keys())
        )
        if state:
            dialog.restoreState(state)
        dialog.setAcceptMode(QFileDialog.AcceptSave)
        if directory:
            dialog.setDirectory(directory)
        ok = dialog.exec_()
        settings.setSaveFileDialogState(dialog.saveState())
        if ok:
            nameFilter = dialog.selectedNameFilter()
            path = dialog.selectedFiles()[0]
            if not os.path.basename(path).endswith(".ufo"):
                path += ".ufo"
            self.saveFile(path, fileFormats[nameFilter])
            app = QApplication.instance()
            app.setCurrentFile(self._font.path)
            self.setWindowTitle(self.fontTitle())
        # return ok

    def reloadFile(self):
        font = self._font
        path = font.path or font.binaryPath
        if not font.dirty or path is None:
            return
        if not ReloadMessageBox.getReloadDocument(self, self.fontTitle()):
            return
        if font.path is not None:
            font.reloadInfo()
            font.reloadKerning()
            font.reloadGroups()
            font.reloadFeatures()
            font.reloadLib()
            font.reloadGlyphs(font.keys())
            font.dirty = False
        else:
            # TODO: we should do this in-place
            font_ = font.__class__().new()
            font_.extract(font.binaryPath)
            self.setFont_(font_)

    def exportFile(self):
        params, ok = ExportDialog.getExportParameters(self, self._font)
        if not ok:
            return
        baseName = params["baseName"]
        directory = params["exportDirectory"]
        compression = set(map(str.lower, params["compression"]))
        for format in map(str.lower, params["formats"]):
            fileName = f"{baseName}.{format}"
            path = os.path.join(directory, fileName)
            try:
                self._font.export(path, format, compression=compression)
            except Exception as e:
                msg = (
                    self.tr("This font’s feature file contains an error.")
                    if isinstance(e, FeatureLibError)
                    else None
                )
                errorReports.showCriticalException(e, message=msg)

    # Edit

    def undo(self):
        widget = self.stackWidget.currentWidget()
        if self.isGlyphTab():
            glyph = widget.activeGlyph()
        else:
            glyph = widget.lastSelectedGlyph()
        glyph.undo()

    def redo(self):
        widget = self.stackWidget.currentWidget()
        if self.isGlyphTab():
            glyph = widget.activeGlyph()
        else:
            glyph = widget.lastSelectedGlyph()
        glyph.redo()

    def cut(self):
        self.copy()
        widget = self.stackWidget.currentWidget()
        if self.isGlyphTab():
            glyph = widget.activeGlyph()
            deleteUISelection(glyph)
        else:
            glyphs = widget.glyphs()
            for index in widget.selection():
                glyph = glyphs[index]
                glyph.clear()

    def copy(self):
        font = self._font
        widget = self.stackWidget.currentWidget()
        clipboard = QApplication.clipboard()
        mimeData = QMimeData()
        if self.isGlyphTab():
            glyph = widget.activeGlyph()
            copyGlyph = glyph.getRepresentation("TruFont.FilterSelection")
            packGlyphs = (copyGlyph,)
        else:
            glyphs = self.glyphCellView.glyphs()
            packGlyphs = (
                glyphs[index] for index in sorted(self.glyphCellView.selection())
            )

        svgGlyphs = []
        pickled = []
        for i, glyph in enumerate(packGlyphs):
            pickled.append(glyph.serialize(blacklist=("name", "unicodes")))

            pen = SVGPathPen(font)
            glyph.draw(pen)
            col = i % 5
            row = i // 5
            g = '<g transform="matrix(1,0,0,-1,{:f},{:f})"><path d="{}"/></g>'.format(
                font.info.unitsPerEm * col,
                font.info.unitsPerEm * row,
                pen.getCommands(),
            )
            svgGlyphs.append(g)

        mimeData.setData("application/x-trufont-glyph-data", pickle.dumps(pickled))

        svg = """\
<?xml version="1.0" standalone="no"?>
<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 20010904//EN"
 "http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd">
<svg version="1.0" xmlns="http://www.w3.org/2000/svg">
%s
</svg>
""" % "\n".join(
            svgGlyphs
        )
        mimeData.setData("image/svg+xml", svg.encode("utf-8"))

        clipboard.setMimeData(mimeData)

    def copyAsComponent(self):
        if self.isGlyphTab():
            pass
        else:
            glyphs = self.glyphCellView.glyphs()
            pickled = []
            for index in self.glyphCellView.selection():
                glyph = glyphs[index]
                componentGlyph = glyph.__class__()
                componentGlyph.width = glyph.width
                component = componentGlyph.instantiateComponent()
                component.baseGlyph = glyph.name
                pickled.append(componentGlyph.serialize())
            clipboard = QApplication.clipboard()
            mimeData = QMimeData()
            mimeData.setData("application/x-trufont-glyph-data", pickle.dumps(pickled))
            clipboard.setMimeData(mimeData)

    def paste(self):
        isGlyphTab = self.isGlyphTab()
        widget = self.stackWidget.currentWidget()
        if isGlyphTab:
            glyphs = (widget.activeGlyph(),)
        else:
            selection = self.glyphCellView.selection()
            glyphs = widget.glyphsForIndexes(selection)
        clipboard = QApplication.clipboard()
        mimeData = clipboard.mimeData()
        if mimeData.hasFormat("application/x-trufont-glyph-data"):
            data = pickle.loads(mimeData.data("application/x-trufont-glyph-data"))
            if len(data) == len(glyphs):
                for pickled, glyph in zip(data, glyphs):
                    if isGlyphTab:
                        pasteGlyph = glyph.__class__()
                        pasteGlyph.deserialize(pickled)
                        # TODO: if we serialize selected state, we don't need
                        # to do this
                        pasteGlyph.selected = True
                        if (
                            len(pasteGlyph)
                            or len(pasteGlyph.components)
                            or len(pasteGlyph.anchors)
                        ):
                            glyph.beginUndoGroup()
                            glyph.holdNotifications()
                            count = len(glyph)
                            pen = glyph.getPointPen()
                            # contours, components
                            pasteGlyph.drawPoints(pen)
                            for contour in glyph[count:]:
                                contour.selected = True
                            # anchors
                            for anchor in pasteGlyph.anchors:
                                glyph.appendAnchor(dict(anchor))
                            # guidelines
                            for guideline in pasteGlyph.guidelines:
                                glyph.appendGuideline(dict(guideline))
                            glyph.releaseHeldNotifications()
                            glyph.endUndoGroup()
                    else:
                        glyph.deserialize(pickled)
            return
        if mimeData.hasFormat("image/svg+xml"):
            if len(glyphs) == 1:
                glyph = glyphs[0]
                try:
                    svgPath = SVGPath.fromstring(mimeData.data("image/svg+xml"))
                except Exception:
                    pass
                else:
                    glyph.beginUndoGroup()
                    if not isGlyphTab:
                        glyph.clear()
                    svgPath.draw(glyph.getPen())
                    glyph.endUndoGroup()
                    return
        if mimeData.hasText():
            if len(glyphs) == 1:
                glyph = glyphs[0]
                otherGlyph = glyph.__class__()
                text = mimeData.text()
                try:
                    readGlyphFromString(text, otherGlyph, otherGlyph.getPointPen())
                except Exception:
                    try:
                        svgPath = SVGPath.fromstring(text)
                        svgPath.draw(otherGlyph.getPen())
                    except Exception:
                        return
                glyph.beginUndoGroup()
                if not isGlyphTab:
                    glyph.clear()
                otherGlyph.drawPoints(glyph.getPointPen())
                glyph.endUndoGroup()

    def selectAll(self):
        widget = self.stackWidget.currentWidget()
        if self.isGlyphTab():
            glyph = widget.activeGlyph()
            if glyph.selected:
                for anchor in glyph.anchors:
                    anchor.selected = True
                for component in glyph.components:
                    component.selected = True
            else:
                glyph.selected = True
        else:
            widget.selectAll()

    def deselect(self):
        widget = self.stackWidget.currentWidget()
        if self.isGlyphTab():
            glyph = widget.activeGlyph()
            for anchor in glyph.anchors:
                anchor.selected = False
            for component in glyph.components:
                component.selected = False
            glyph.selected = False
        else:
            widget.setSelection(set())

    def delete(self):
        modifiers = QApplication.keyboardModifiers()
        widget = self.stackWidget.currentWidget()
        if self.isGlyphTab():
            glyph = widget.activeGlyph()
            # TODO: fuse more the two methods, they're similar and delete is
            # Cut except not putting in the clipboard
            if modifiers & Qt.AltModifier:
                deleteUISelection(glyph)
            else:
                preserveShape = not modifiers & Qt.ShiftModifier
                removeUIGlyphElements(glyph, preserveShape)
        else:
            erase = modifiers & Qt.ShiftModifier
            if self._proceedWithDeletion(erase):
                glyphs = widget.glyphsForIndexes(widget.selection())
                for glyph in glyphs:
                    font = glyph.font
                    for layer in font.layers:
                        if glyph.name in layer:
                            defaultLayer = layer[glyph.name] == glyph
                            if defaultLayer and not erase:
                                # TODO: clear in glyph.template setter?
                                glyph.clear()
                                glyph.template = True
                            else:
                                del layer[glyph.name]

    def findGlyph(self):
        widget = self.stackWidget.currentWidget()
        if self.isGlyphTab():
            glyph = widget.activeGlyph()
            newGlyph, ok = FindDialog.getNewGlyph(self, glyph)
            if ok and newGlyph is not None:
                widget.setActiveGlyph(newGlyph)
        else:
            pass  # XXX

    # View

    def zoom(self, step):
        if self.isGlyphTab():
            widget = self.stackWidget.currentWidget()
            newScale = widget.scale() * pow(1.2, step)
            widget.zoom(newScale)
            self.statusBar.setSize(widget.pointSize())
        else:
            value = self.statusBar.size()
            newValue = value + 10 * step
            self.statusBar.setSize(newValue)

    def resetZoom(self):
        widget = self.stackWidget.currentWidget()
        if self.isGlyphTab():
            widget.fitScaleBBox()
        else:
            settings.removeGlyphCellSize()
            cellSize = settings.glyphCellSize()
            self.statusBar.setSize(cellSize)

    def tabOffset(self, value):
        tab = self.tabWidget.currentTab()
        newTab = (tab + value) % len(self.tabWidget.tabs())
        self.tabWidget.setCurrentTab(newTab)

    def glyphOffset(self, value):
        widget = self.stackWidget.currentWidget()
        if self.isGlyphTab():
            currentGlyph = widget.activeGlyph()
            font = currentGlyph.font
            glyphOrder = font.glyphOrder
            # should be enforced in fontView already
            if not (glyphOrder and len(glyphOrder)):
                return
            index = glyphOrder.index(currentGlyph.name)
            newIndex = (index + value) % len(glyphOrder)
            glyph = font[glyphOrder[newIndex]]
            widget.setActiveGlyph(glyph)
        else:
            lastSelectedCell = widget.lastSelectedCell()
            if lastSelectedCell is None:
                return
            newIndex = lastSelectedCell + value
            if newIndex < 0 or newIndex >= len(widget.glyphs()):
                return
            widget.setSelection({newIndex})

    def layerOffset(self, value):
        widget = self.stackWidget.currentWidget()
        if self.isGlyphTab():
            currentGlyph = widget.activeGlyph()
            layerSet, layer = currentGlyph.layerSet, currentGlyph.layer
            if None in (layerSet, layer):
                return
            index = layerSet.layerOrder.index(layer.name)
            newIndex = (index + value) % len(layerSet)
            layer_ = layerSet[layerSet.layerOrder[newIndex]]
            if layer_ == layer:
                return
            # XXX: fix get
            # glyph = layer_.get(currentGlyph.name)
            if currentGlyph.name in layer_:
                glyph = layer_[currentGlyph.name]
            else:
                glyph = layer_.newGlyph(currentGlyph.name)
            widget.setActiveGlyph(glyph)

    # Font

    def fontInfo(self):
        # If a window is already opened, bring it to the front, else spawn one.
        # TODO: see about using widget.setAttribute(Qt.WA_DeleteOnClose)
        # otherwise it seems we're just leaking memory after each close...
        # (both raise_ and show allocate memory instead of using the hidden
        # widget it seems)
        if self._infoWindow is not None and self._infoWindow.isVisible():
            self._infoWindow.raise_()
        else:
            self._infoWindow = FontInfoWindow(self._font, self)
            self._infoWindow.show()

    def fontFeatures(self):
        # TODO: see up here
        if self._featuresWindow is not None and self._featuresWindow.isVisible():
            self._featuresWindow.raise_()
        else:
            self._featuresWindow = FontFeaturesWindow(self._font, self)
            self._featuresWindow.show()

    def addGlyphs(self):
        glyphs = self.glyphCellView.glyphs()
        newGlyphNames, params, ok = AddGlyphsDialog.getNewGlyphNames(self, glyphs)
        if ok:
            sortFont = params.pop("sortFont")
            for name in newGlyphNames:
                glyph = self._font.get(name, **params)
                if glyph is not None:
                    glyphs.append(glyph)
            self.glyphCellView.setGlyphs(glyphs)
            if sortFont:
                # TODO: when the user add chars from a glyphSet and no others,
                # should we try to sort according to that glyphSet?
                # The above would probably warrant some rearchitecturing.
                # kick-in the sort mechanism
                self._font.sortDescriptor = self._font.sortDescriptor

    def sortGlyphs(self):
        sortDescriptor, ok = SortDialog.getDescriptor(self, self._font.sortDescriptor)
        if ok:
            self._font.sortDescriptor = sortDescriptor

    # Window

    def groups(self):
        # TODO: see up here
        if self._groupsWindow is not None and self._groupsWindow.isVisible():
            self._groupsWindow.raise_()
        else:
            self._groupsWindow = GroupsWindow(self._font, self)
            self._groupsWindow.show()

    def kerning(self):
        # TODO: see up here
        if self._kerningWindow is not None and self._kerningWindow.isVisible():
            self._kerningWindow.raise_()
        else:
            self._kerningWindow = KerningWindow(self._font, self)
            self._kerningWindow.show()

    def metrics(self):
        # TODO: see up here
        if self._metricsWindow is not None and self._metricsWindow.isVisible():
            self._metricsWindow.raise_()
        else:
            self._metricsWindow = MetricsWindow(self._font)
            # XXX: need proper, fast windowForFont API!
            self._metricsWindow._fontWindow = self
            self.destroyed.connect(self._metricsWindow.close)
            self._metricsWindow.show()
        # TODO: default string kicks-in on the window before this. Figure out
        # how to make a clean interface
        selection = self.glyphCellView.selection()
        if selection:
            glyphs = self.glyphCellView.glyphsForIndexes(selection)
            self._metricsWindow.setGlyphs(glyphs)

    def properties(self):
        shouldBeVisible = self.propertiesView.isHidden()
        self.propertiesView.setVisible(shouldBeVisible)
        self.writeSettings()

    # update methods

    def _setGlyphPreview(self, value):
        index = self.stackWidget.currentIndex()
        if index:
            widget = self.stackWidget.currentWidget()
            widget.setPreviewEnabled(value)

    def _updateCurrentGlyph(self):
        # TODO: refactor this pattern...
        widget = self.stackWidget.currentWidget()
        if self.isGlyphTab():
            glyph = widget.activeGlyph()
        else:
            glyph = widget.lastSelectedGlyph()
        if glyph is not None:
            app = QApplication.instance()
            app.setCurrentGlyph(glyph)

    def _updateGlyphActions(self):
        if not hasattr(self, "_undoAction"):
            return
        widget = self.stackWidget.currentWidget()
        if self.isGlyphTab():
            currentGlyph = widget.activeGlyph()
        else:
            currentGlyph = widget.lastSelectedGlyph()
        # disconnect eventual signal of previous glyph
        objects = ((self._undoAction, self.undo), (self._redoAction, self.redo))
        for action, slot in objects:
            try:
                action.disconnect()
            except TypeError:
                pass
            action.triggered.connect(slot)
        # now update status
        if currentGlyph is None:
            self._undoAction.setEnabled(False)
            self._redoAction.setEnabled(False)
        else:
            undoManager = currentGlyph.undoManager
            self._undoAction.setEnabled(currentGlyph.canUndo())
            undoManager.canUndoChanged.connect(self._undoAction.setEnabled)
            self._redoAction.setEnabled(currentGlyph.canRedo())
            undoManager.canRedoChanged.connect(self._redoAction.setEnabled)
        # and other actions
        for action in self._clipboardActions:
            action.setEnabled(currentGlyph is not None)

    # helper

    def _proceedWithDeletion(self, erase=False):
        if not self.glyphCellView.selection():
            return
        tr = self.tr("Delete") if erase else self.tr("Clear")
        text = self.tr("Do you want to %s selected glyphs?") % tr.lower()
        closeDialog = QMessageBox(
            QMessageBox.Question,
            "",
            self.tr("%s glyphs") % tr,
            QMessageBox.Yes | QMessageBox.No,
            self,
        )
        closeDialog.setInformativeText(text)
        closeDialog.setModal(True)
        ret = closeDialog.exec_()
        if ret == QMessageBox.Yes:
            return True
        return False

    # ----------
    # Qt methods
    # ----------

    def setWindowTitle(self, title):
        if platformSpecific.appNameInTitle():
            title += " – TruFont"
        super().setWindowTitle(f"[*]{title}")

    def sizeHint(self):
        return QSize(1270, 800)

    def moveEvent(self, event):
        self.writeSettings()

    resizeEvent = moveEvent

    def showEvent(self, event):
        app = QApplication.instance()
        data = dict(font=self._font, window=self)
        app.postNotification("fontWindowWillOpen", data)
        super().showEvent(event)
        app.postNotification("fontWindowOpened", data)

    def closeEvent(self, event):
        ok = self.maybeSaveBeforeExit()
        if ok:
            app = QApplication.instance()
            data = dict(font=self._font, window=self)
            app.postNotification("fontWindowWillClose", data)
            self._font.removeObserver(self, "Font.Changed")
            app = QApplication.instance()
            app.dispatcher.removeObserver(self, "drawingToolRegistered")
            app.dispatcher.removeObserver(self, "drawingToolUnregistered")
            app.dispatcher.removeObserver(self, "glyphViewGlyphsChanged")
            event.accept()
        else:
            event.ignore()

    def event(self, event):
        if event.type() == QEvent.WindowActivate:
            app = QApplication.instance()
            app.setCurrentFontWindow(self)
            self._updateCurrentGlyph()
        return super().event(event)

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.fillRect(event.rect(), QColor(212, 212, 212))



from defcon import Glyph, Image, registerRepresentationFactory

from defconQt.representationFactories.glyphCellFactory import GlyphCellFactory
from defconQt.representationFactories.glyphViewFactory import (
    NoComponentsQPainterPathFactory,
    OnlyComponentsQPainterPathFactory,
    OutlineInformationFactory,
    QPixmapFactory,
)
from defconQt.representationFactories.qPainterPathFactory import QPainterPathFactory

# TODO: fine-tune the destructive notifications
_glyphFactories = {
    "defconQt.QPainterPath": (QPainterPathFactory, None),
    "defconQt.NoComponentsQPainterPath": (NoComponentsQPainterPathFactory, None),
    "defconQt.OnlyComponentsQPainterPath": (OnlyComponentsQPainterPathFactory, None),
    "defconQt.GlyphCell": (GlyphCellFactory, None),
    "defconQt.OutlineInformation": (
        OutlineInformationFactory,
        ("Glyph.Changed", "Glyph.SelectionChanged"),
    ),
}
_imageFactories = {
    "defconQt.QPixmap": (
        QPixmapFactory,
        ("Image.FileNameChanged", "Image.ColorChanged", "Image.ImageDataChanged"),
    )
}


def registerAllFactories():
    for name, (factory, destructiveNotifications) in _glyphFactories.items():
        registerRepresentationFactory(
            Glyph, name, factory, destructiveNotifications=destructiveNotifications
        )
    for name, (factory, destructiveNotifications) in _imageFactories.items():
        registerRepresentationFactory(
            Image, name, factory, destructiveNotifications=destructiveNotifications
        )



from PyQt5.QtCore import QRectF, QSize, Qt
from PyQt5.QtGui import QColor, QPainter, QPainterPath
from PyQt5.QtWidgets import QSizePolicy, QWidget


class GlyphAlignmentWidget(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)

        self._alignment = None
        self._alignmentPaths = []
        self._glyph = None

        self._color = QColor(130, 130, 130)
        self._selectedColor = QColor(20, 146, 230)

    # synthetized properties

    def hOrigin(self):
        alignment = self._alignment
        if alignment is not None:
            if not alignment % 3:
                return 0
            elif not (alignment - 2) % 3:
                return 2
        return 1

    def vOrigin(self):
        alignment = self._alignment
        if alignment is not None:
            if alignment < 3:
                return 2
            elif alignment > 5:
                return 0
        return 1

    def origin(self, representation=None):
        alignment = self._alignment
        glyph = self._glyph
        if alignment is None or not glyph:
            return (0, 0)
        if representation is not None:
            glyph = glyph.getRepresentation(representation)
            if not glyph:
                return (0, 0)
        left, bottom, right, top = glyph.controlPointBounds
        if not alignment % 3:
            x = left
        elif not (alignment - 2) % 3:
            x = right
        else:
            x = (left + right) / 2
        if alignment < 3:
            y = top
        elif alignment > 5:
            y = bottom
        else:
            y = (top + bottom) / 2
        return (x, y)

    # properties

    def alignment(self):
        return self._alignment

    def setAlignment(self, value):
        self._alignment = value
        self.update()

    def glyph(self):
        return self._glyph

    def setGlyph(self, glyph):
        self._glyph = glyph

    def color(self):
        return self._color

    def setColor(self, color):
        self._color = color
        self.update()

    def selectedColor(self):
        return self._selectedColor

    def setSelectedColor(self, color):
        self._selectedColor = color
        self.update()

    # ----------
    # Qt methods
    # ----------

    def sizeHint(self):
        return QSize(27, 27)

    def mousePressEvent(self, event):
        if event.button() & Qt.LeftButton:
            pos = event.localPos()
            # TODO: press should be actuated on release...
            for index, path in enumerate(self._alignmentPaths):
                if path.contains(pos):
                    if self._alignment == index:
                        self._alignment = None
                    else:
                        self._alignment = index
                    self.update()
                    break
        else:
            super().mousePressEvent(event)

    def paintEvent(self, event):
        self._alignmentPaths = []
        painter = QPainter(self)
        painter.setPen(self._color)

        circleRadius = 2.5
        selectedRadius = 3
        padding = 1
        rect = event.rect()
        size = min(rect.height(), rect.width())
        offset = round(0.5 * (rect.width() - size))
        painter.translate(offset, 0)
        borderRect = rect.__class__(
            rect.left() + circleRadius + padding,
            rect.top() + circleRadius + padding,
            size - 2 * (circleRadius + padding),
            size - 2 * (circleRadius + padding),
        )
        borderPath = QPainterPath()
        borderPath.addRect(*borderRect.getRect())

        columnCount = 3
        radioPath = QPainterPath()
        selectedPath = QPainterPath()
        for row in range(columnCount):
            for col in range(columnCount):
                index = row * columnCount + col
                x, y = (
                    padding + col * 0.5 * borderRect.width(),
                    padding + row * 0.5 * borderRect.height(),
                )
                delta = selectedRadius - circleRadius
                sx = x - delta
                sy = y - delta
                if self._alignment == index:
                    path = QPainterPath()
                    path.addEllipse(sx, sy, 2 * selectedRadius, 2 * selectedRadius)
                    selectedPath = path
                else:
                    path = QPainterPath()
                    path.addEllipse(x, y, 2 * circleRadius, 2 * circleRadius)
                    radioPath.addPath(path)
                self._alignmentPaths.append(
                    QRectF(sx + offset, sy, 2 * selectedRadius, 2 * selectedRadius)
                )
        painter.drawPath(borderPath - radioPath)
        painter.setRenderHint(QPainter.Antialiasing)
        painter.fillPath(radioPath, self._color)
        painter.fillPath(selectedPath, self._selectedColor)
